"""Collapse reads into their underlying molecules.

This module contains functions for the collapse and error correction of MPX data
(from FASTQ)

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
import os
import tempfile
import typing
import warnings
from collections import Counter, defaultdict
from typing import (
    Generator,
    Iterable,
    Iterator,
    Literal,
    Optional,
)

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyfastx
from annoy import AnnoyIndex

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", module="umi_tools")
    from umi_tools._dedup_umi import edit_distance
    from umi_tools.network import breadth_first_search

from pathlib import Path

from pixelator.common.utils import gz_size
from pixelator.mpx.collapse.constants import SEED

logger = logging.getLogger("pixelator.mpx.collapse")

np.random.seed(SEED)

UniqueFragment = str
UpiB = str
UniqueFragmentToUpiB = dict[UniqueFragment, list[UpiB]]


class FileFqGzEmpty(Exception):
    """Class to manage empty fastq.gz file exceptions.

    Attributes
    ----------
        msg: the error message to output
        fname: the name of the file
        size: the size of the file uncompressed (should be 0)

    """

    def __init__(self, msg: str, fname: str | Path, size: int):
        """Initialize the exception."""
        self.msg = msg
        self.fname = os.fspath(fname)
        self.size = size


class CollapsedFragment(typing.NamedTuple):
    """A collapsed fragment.

    :attr sequence: the consensus sequence for a list of similar fragments
    :attr unique_molecules_count: the number of unique fragments that are
        represented by this collapsed fragment.
    :attr read_count: the number of reads that are represented
        by this collapsed fragment
    """

    sequence: str
    unique_molecules_count: int
    reads_count: int


def build_annoytree(data: npt.NDArray[np.uint8], n_trees: int = 10) -> AnnoyIndex:
    """Build an Annoy tree index.

    Create an `Annoy` tree [1]_ using the given data matrix (an array of DNA sequences
    in binary format) and using hamming distance as the distance metric in the Annoy
    index.

    The number of trees (`n_trees`) determine the accuracy of the nearest neighbour
    estimations.

    .. [1] https://github.com/spotify/annoy

    :param data: the array of sequences (n_ele, n_features) in binary format
    :param n_trees: the number of Annoy trees to build. Default: 10
    :returns: an AnnoyIndex object
    :rtype: AnnoyIndex
    """
    logger.debug("Building annoy tree of shape %i,%i", data.shape[0], data.shape[1])

    tree = AnnoyIndex(data.shape[1], "hamming")
    tree.set_seed(SEED)
    for i in range(data.shape[0]):
        tree.add_item(i, data[i, :])
    tree.build(n_trees, n_jobs=-1)

    logger.debug("Annoy tree built")
    return tree


def build_binary_data(seqs: list[str]) -> npt.NDArray[np.uint8]:
    """Build a binary matrix from a list of sequences using two-bit encoding.

    Convert a list of DNA sequences (str) to binary sequences
    using two bit encoding, the converted sequences are returned
    as an `np.array` (n_ele, n_features).

    :param seqs: a list of DNA sequences
    :returns: a numpy array of binary sequences
    :rtype: npt.NDArray[np.uint8]
    """
    logger.debug("Transforming %i sequences to binary form", len(seqs))

    tr = {"A": "00", "C": "01", "G": "10", "T": "11"}
    n_feat = len(seqs[0])
    n_ele = len(seqs)
    data = np.zeros((n_ele, n_feat * 2), np.uint8)
    for i, seq in enumerate(seqs):
        data[i] = [int(b) for s in seq for b in tr[s]]

    logger.debug("Sequences transformed")
    return data


def get_collapsed_fragments_for_component(  # noqa: DOC402,DOC404
    components: list[set[UniqueFragment]], counts: dict[UniqueFragment, int]
) -> Generator[CollapsedFragment, None, None]:
    """Take the representative sequence from a component based on its counts.

    Given a list of components (i.e. sequences that presumably belong to the
    same original molecule), and the counts of occurrences for each of these,
    return an iterator of objects with the representative sequence for that
    component the total number of reads associated with the cluster and
    the number of collapsed fragments.

    The representative molecule is selected as the one with the most associated
    upib's, and if there is a tie the lexicographically smallest sequence is picked
    to make the results reproducible between runs.

    :param components: a list of components as produced from `get_connected_components`
    :param counts: a dictionary of the counts of each unique fragment
    :rtype: Generator[CollapsedFragment, None, None]
    :yields CollapsedFragment: a collapsed fragment
    """
    for component in components:
        # get the most common sequence in the component and use
        # this as the 'consensus sequence'

        # get an iterator over the fragments
        fragment_it = iter(component)

        # Initialize variables for finding the representative with the first fragment
        # of a component
        representative_frag = next(fragment_it)
        representative_frag_count = counts[representative_frag]
        total_reads = representative_frag_count

        # Loop over the other fragments to find the representative
        for fragment in fragment_it:
            fragment_read_counts = counts[fragment]
            total_reads += fragment_read_counts
            if fragment_read_counts > representative_frag_count or (
                fragment_read_counts == representative_frag_count
                and fragment < representative_frag
            ):
                representative_frag = fragment
                representative_frag_count = fragment_read_counts

        collapsed_fragments = len(component)
        yield CollapsedFragment(representative_frag, collapsed_fragments, total_reads)


# This code snippet has been obtained from:
# https://github.com/CGATOxford/umi-tools
def get_connected_components(
    graph: dict[UniqueFragment, list[UniqueFragment]], counts: dict[UniqueFragment, int]
) -> list[set[UniqueFragment]]:
    """Get connected components from a graph of sequences.

    Find the connected components from the sequence graph (represented by an adjacency
    dictionary). Each node in the graph is a sequence and each edge represents that
    those sequences similar (e.g. have a low edit distance between each other). Thus,
    a connected component in this graph can be seen as a group of sequences deriving
    from the same original molecule.

    The function will return the list of connected components of each sequence
    (not counting sequences twice).

    :param graph: a dictionary of sequence to sequences (within same distance)
    :param counts: a dictionary of counts (copies) for each sequence in `graph`
    :returns: a list of sets of sequences where each set represents a
              connected component
    :rtype: list[set[UniqueFragment]]
    """
    found = set()
    components = []
    for node in sorted(graph, key=lambda x: counts[x], reverse=True):
        if node not in found:
            component = breadth_first_search(node, graph)
            found.update(component)
            components.append(component)
    return components


def identify_fragments_to_collapse(
    seqs: list[UniqueFragment],
    min_dist: int,
    max_neighbours: int,
) -> dict[UniqueFragment, list[UniqueFragment]]:
    """Identify fragments to collapse by approximate nearest neighbourhood search.

    Tries to identify all sequences in the given list of sequences
    (`seqs`) within a hamming distance `min_dist` of each other, and group them
    together.

    An `Annoy` index (approximate nearest neighbour) is used to avoid having to
    compute the distances between all sequence pairs. This means that this function is
    not guaranteed to return all sequences within a given hamming distance.

    The parameter `max_neighbours` is used to control the maximum number of
    neighbours to look for. The higher this number the more likely you are
    to find all sequences within a hamming distance of `min_dist`, but the search
    will also be slower. The number of actual neighbours searched is determined as
    follows: pick the minimum number of `max_neighbours` and 10% of the input data,
    but at least search 10 neighbours of each sequence.

    :param seqs: a list of sequences to be grouped
    :param min_dist: the hamming distance threshold (i.e. the mismatches
                     between two sequences)
    :param max_neighbours: the number of neighbours to use in the Annoy index
    :returns: a dictionary with fragments as keys and a list of their adjoining
              fragments as values
    :rtype: dict[UniqueFragment, list[UniqueFragment]]
    """
    logger.debug("Computing adjacency sequences from %i elements", len(seqs))

    # use a nearest neighbours tree to reduce the search space
    # TODO this approach requires memory (n_ele * len(seq) * 2)
    # we could add an alternative approach using on binary data
    # but this has proven to be slower
    data = build_binary_data(seqs)
    tree = build_annoytree(data)

    neighbours = max(10, max_neighbours)
    logger.debug("The number of neighbours was set to %s", neighbours)

    # iterate the neighbours of each sequence to obtain
    # the adjacency list (dictionary of lists)
    adj_list: dict[str, list[str]] = {seq: [] for seq in seqs}
    for i, seq1 in enumerate(seqs):
        if i % 100000 == 0:
            logger.debug("Processed 100,000 reads...")
        idxs, estimated_distances = tree.get_nns_by_item(
            i, n=neighbours, search_k=-1, include_distances=True
        )
        for idx, dist in zip(idxs, estimated_distances):
            seq2 = seqs[idx]

            # No need to add yourself to the adjacency list
            if seq1 == seq2:
                continue

            # Escape early if estimated distance is more than twice the
            # minimum distance allowed, since there is a very low risk
            # that it will be a dist < min distance when we calculate the
            # true edit distance below.
            # This works since the distances returned from Annoy above
            # are returned on order of lowest to highest distance.
            # Please not that since we are using Hamming distance,
            # the min_dist * 2 is the equivalent of correcting min_dist * 2
            # base pairs.
            if dist > min_dist * 2:
                break

            # The`dist` provided by Annoy above is only an approximation of the
            # distance between the sequences. Because of this we need to double
            # check it here
            if edit_distance(seq1.encode("utf-8"), seq2.encode("utf-8")) <= min_dist:
                adj_list[seq1].append(seq2)

    logger.debug("Approximate sequence adjacency computed")
    return adj_list


def collapse_sequences_unique(
    seq_dict: UniqueFragmentToUpiB,
) -> Generator[CollapsedFragment, None, None]:
    """Get all fragments.

    Let each key in `seq_dict` represent it's own sequence. This is equivalent
    to not collapsing the sequences.

    :param seq_dict: the fragment to upib dict
    :yield a CollapsedFragment object
    """
    logger.debug("Picking all unique sequences (i.e. no collapsing is carried out)")

    for seq in seq_dict.keys():
        yield CollapsedFragment(seq, 1, len(seq_dict[seq]))


def collapse_sequences_adjacency(
    seq_dict: UniqueFragmentToUpiB,
    max_neighbours: int,
    min_dist: int,
) -> Iterator[CollapsedFragment]:
    """Collapse sequences based on their adjacency.

    Tries to identify all fragments that represent the same underlying
    molecule by trying to find sequences that are less than or equal to `min_distance`
    apart. The distance measure used is hamming distance.

    Each original collapsed molecule is represented by the fragment that is associated
    with the largest number of upib's.

    It does so by doing approximate nearest neighbour estimation, and then checking
    the exact distance between the neighbours. For more information on this see
    `identify_fragments_to_collapse`.

    :param seq_dict: a dictionary mapping unique fragments to their
                     corresponding upib's
    :param max_neighbours: the maximum number of neighbours to search in the approximate
                           nearest neighbour search
    :param min_dist: the hamming distance threshold (i.e. the mismatches
                     between two sequences)
    :returns: An iterator of the of collapsed molecules, and their original counts
    :rtype: Iterator[CollapsedFragment]
    """
    logger.debug("Collapsing %i sequences", len(seq_dict))

    counts = {k: len(v) for k, v in seq_dict.items()}
    seqs = list(counts.keys())

    adj_list = identify_fragments_to_collapse(
        seqs=seqs, max_neighbours=max_neighbours, min_dist=min_dist
    )
    full_components = get_connected_components(adj_list, counts)
    components = get_collapsed_fragments_for_component(full_components, counts)
    return components


def create_fragment_to_upib_dict(
    input_file: str,
    upia_start: int,
    upia_end: int,
    upib_start: int,
    upib_end: int,
    umia_start: Optional[int] = None,
    umia_end: Optional[int] = None,
    umib_start: Optional[int] = None,
    umib_end: Optional[int] = None,
) -> tuple[UniqueFragmentToUpiB, int]:
    """Create a dict mapping fragments to upib's.

    Parses a fastq file with pixel data and extracts the UPIA, UPIB,
    UMIA and UMIB and then returns a dictionary of UMI+UPIA -> [UPIB].

    UMIA will be ignored if any of its positions is None.
    UMIB will be ignored if any of its positions is None.

    :param input_file: path to the file to read
    :param upia_start: the 0-based start position of UPIA
    :param upia_end: the 1-based end position of UPIA
    :param upib_start: the 0-based start position of UPIB
    :param upib_end: the 1-based end position of UPIB
    :param umia_start: the 0-based start position of UMIA
    :param umia_end: the 1-based end position of UMIA
    :param umib_start: the 0-based start position of UMIB
    :param umib_end: the 1-based end position of UMIB
    :returns: a tuple with:
        - a dictionary with the sequence of umi+upia as keys
          and the list of associated upibs as values.
        - the number of input reads
    :rtype: tuple[UniqueFragmentToUpiB, int]
    :raises FileFqGzEmpty: when the file is empty
    :raises RuntimeError: when there is a error parsing the file
    """
    logger.debug("Extracting umi-upi sequences from %s", input_file)

    has_umia = umia_start is not None and umia_end is not None
    has_umib = umib_start is not None and umib_end is not None
    seqs_dict = defaultdict(list)
    input_reads_count = 0

    try:
        for _, seq, _ in pyfastx.Fastq(input_file, build_index=False):
            upia = seq[upia_start:upia_end]
            upib = seq[upib_start:upib_end]
            if has_umia and has_umib:
                umi = seq[umia_start:umia_end] + seq[umib_start:umib_end]
            elif has_umia:
                umi = seq[umia_start:umia_end]
            elif has_umib:
                umi = seq[umib_start:umib_end]
            else:
                umi = ""
            seq = umi + upia
            seqs_dict[seq].append(upib)
            input_reads_count += 1
    except Exception as exc:
        gzlen = gz_size(input_file)
        if gzlen == 0:
            raise FileFqGzEmpty(
                f"File {input_file} is empty", input_file, gzlen
            ) from exc
        raise RuntimeError(
            f"There was an error '{str(exc)}' parsing {input_file}"
        ) from exc

    logger.debug("umi-upi sequences were extracted from %s", input_file)
    return seqs_dict, input_reads_count


def filter_by_minimum_upib_count(
    unique_reads: UniqueFragmentToUpiB, min_count: int
) -> UniqueFragmentToUpiB:
    """Filter fragment to upib's dictionary by minimum upib count.

    Filter reads from the input dictionary requiring at least `min_count` or more
    instances of the upib's to keep it.

    :param unique_reads: a dictionary of fragments and their corresponding upib's
    :param min_count: the minimum number of upib's per fragment required to keep it
    :returns: A filtered instance of the input dictionary
    :rtype: UniqueFragmentToUpiB
    """
    unique_reads = {k: v for k, v in unique_reads.items() if len(v) >= min_count}
    # in case there are no reads after filtering
    return unique_reads


def create_edgelist(
    clustered_reads: Iterable[CollapsedFragment],
    unique_reads: UniqueFragmentToUpiB,
    umia_start: Optional[int],
    umia_end: Optional[int],
    umib_start: Optional[int],
    umib_end: Optional[int],
    marker: str,
    sequence: str,
) -> pd.DataFrame:
    """Create an edgelist.

    Create an egdelist of the MPX graph based on the unique reads found,
    and their clustering.

    `clustered_reads` should be an iterable where each element is a tuple
    of reads that belong to the same cluster, i.e. are derived from the same
    unique fragment.

    `unique_reads` should contain a mapping from each of these unique fragments
    to their corresponding upib's.

    This will create an edgelist where each unique fragment is a row, with information
    about:
        - `upia`, the upia of the fragment
        - `upib`, the upib of the fragment
        - `umi`, the umi of the fragment
        - `count`, the number of upib's associated with the fragment
        - `unique_molecules_count`, the number of unique molecules (based on upia+umi)
           associated with the collapsed molecule
        - `marker`, the marker associated with this fragment
        - `sequence`, the antibody DNA-oligo sequence of the marker associated
           with the fragment

    :param clustered_reads: An iterable of tuples of unique fragments
    :param unique_reads: A UniqueFragmentToUpiB dictionary
    :param umia_start: the start position of upia
    :param umia_end: the end position of upia
    :param umib_start: the start position of upib
    :param umib_end: the end position of upib
    :param marker: the marker
    :param sequence: the sequence of the marker:
    :returns: a dataframe representing the edgelist of the mpx graph
    :rtype: pd.DataFrame
    """
    # get the umi sizes to do the split
    if umia_start is not None and umia_end is not None:
        umia_size = umia_end - umia_start
    else:
        umia_size = 0
    if umib_start is not None and umib_end is not None:
        umib_size = umib_end - umib_start
    else:
        umib_size = 0
    umi_size = umia_size + umib_size

    # iterate each collapsed umi+upia to get the upib
    # with the highest count and store in a list
    def data():
        for (
            cluster_representative_fragment,
            unique_molecules_count,
            read_count,
        ) in clustered_reads:
            # get all the upis from the sequence in the cluster
            upibs = unique_reads[cluster_representative_fragment]

            # The cluster_representative_fragment is the concatenation of UMI + UPIA
            umi = cluster_representative_fragment[:umi_size]
            upia = cluster_representative_fragment[umi_size:]

            # take the most common upib from the list
            unique_upibs = Counter(upibs)
            upib, _ = unique_upibs.most_common(1)[0]

            # update data array
            yield (upia, upib, umi, read_count, unique_molecules_count)

    # create an edge list (pd.DataFrame) with the collapsed sequences
    df = pd.DataFrame(
        data=data(),
        columns=["upia", "upib", "umi", "count", "unique_molecules_count"],
    )
    df.insert(3, "marker", marker)
    df.insert(4, "sequence", sequence)

    return df


def write_tmp_feather_file(df: pd.DataFrame) -> Path:
    """Write the dataframe to a feather file in the OS tmpdir.

    :param df: the data frame to write
    :returns: path of the file written
    :rtype: PathType
    """
    # create a temporary edge list and save it to a temp file
    tmp_file = tempfile.mkstemp(suffix=".feather")[1]
    df.to_feather(tmp_file)
    return Path(tmp_file)


def collapse_fastq(
    input_file: str,
    algorithm: Literal["unique", "adjacency"],
    marker: str,
    sequence: str,
    upia_start: int,
    upia_end: int,
    upib_start: int,
    upib_end: int,
    umia_start: Optional[int],
    umia_end: Optional[int],
    umib_start: Optional[int],
    umib_end: Optional[int],
    max_neighbours: Optional[int] = None,
    mismatches: Optional[int] = None,
    min_count: Optional[int] = None,
) -> tuple[Path | None, int]:
    """Collapses reads from a fastq file.

    Takes a fastq file as input and collapses its
    reads by UMI + UPIA (first) to then take the most abundant
    UPIB. The collapsed (error corrected) reads are saved as a
    `pd.DataFrame` to a temporary file with the following columns:

    upia,upib,umi,marker,count

    The function returns the path to the file or None if the input
    fastq file is empty or corrupted.

    The location (position) of the UPIs and UMIs must be provided.

    The function uses some of the functionalities from `umi-tools` [1]_.

    When `algorithm` is `unique` only the unique sequences will be returned
    (no error correction), otherwise sequences will be collapsed based
    on a hamming distance (`mismatches`), using an approximate nearest
    neighbour search based on Annoy [2]_.

    .. [1] Smith, Tom, Andreas Heger, and Ian Sudbery. 2017. “UMI-Tools: Modeling
           Sequencing Errors in Unique Molecular Identifiers to Improve
           Quantification Accuracy.” Genome Research 27 (3): 491–99.

    .. [2] https://github.com/spotify/annoy

    :param input_file: the path to the fastq file containing MPX amplicons
                       (must contain the UPI and UMI)
    :param algorithm: the collapsing algorithm to use (unique or adjacency)
    :param marker: the antibody tag to append to the output
    :param sequence: the barcode sequence to append to the output
    :param upia_start: the start position (0-based) of UPIA
    :param upia_end: the end position (1-based) of UPIA
    :param upib_start: the start position (0-based) of UPIB
    :param upib_end: the end position (1-based) of UPIB
    :param umia_start: the start position (0-based) of UMIA
                       (if None UMIA will be ignored)
    :param umia_end: the end position (1-based) of UMIA
                     (if None UMIA will be ignored)
    :param umib_start: the start position (0-based) of UMIB
                      (if None UMIB will be ignored)
    :param umib_end: the end position (1-based) of UMIB
                     (if None UMIB will be ignored)
    :param max_neighbours: the number of neighbours used in the approximate nearest
                           neighbour search
    :param mismatches: the number of mismatches allowed between sequences
    :param min_count: discard reads with a count lower than this
    :returns: a str containing the path to the edge list file
    :rtype: tuple[Path | None, int]
    :raises AssertionError: invalid input
    :raises RuntimeError: raises an exception
    """
    if algorithm not in ["unique", "adjacency"]:
        raise AssertionError(f"Invalid value {algorithm} for algorithm")

    if algorithm == "adjacency":
        if not max_neighbours or not mismatches:
            raise AssertionError(
                (
                    'When `algorithm` is "adjacency", `max_neighbours`'
                    "and `mismatches` must be set'"
                )
            )

    logger.debug("Collapsing reads from %s", input_file)

    # parse input fastq file to extract a dict of umi+upia -> [upib]
    # TODO: this dictionary is memory heavy, a possible solution is to
    # replace it by a disk-based dictionary but that will really affect
    # the running time. Another approach would be to create a class with
    # a look up table and store the sequences in a list of tuples
    try:
        unique_reads, input_reads_count = create_fragment_to_upib_dict(
            input_file,
            upia_start,
            upia_end,
            upib_start,
            upib_end,
            umia_start,
            umia_end,
            umib_start,
            umib_end,
        )
    except FileFqGzEmpty as err:
        # we want to allow empty files
        logger.warning(str(err.msg))
        return (None, 0)

    # In case there are no reads in the input file
    if not unique_reads:
        # we want to allow empty files
        logger.warning("The input file %s does not contain data", input_file)
        return (None, 0)

    if min_count and min_count > 1:
        unique_reads = filter_by_minimum_upib_count(unique_reads, min_count)
        if not unique_reads:
            logger.warning(
                (
                    "The input file %s does not any contain"
                    "reads after filtering by count >= %i"
                ),
                input_file,
                min_count,
            )
            return (None, input_reads_count)

    if algorithm == "adjacency":
        clustered_reads = collapse_sequences_adjacency(
            seq_dict=unique_reads,
            max_neighbours=max_neighbours,  # type: ignore
            min_dist=mismatches,  # type: ignore
        )
    elif algorithm == "unique":
        clustered_reads = collapse_sequences_unique(
            seq_dict=unique_reads,
        )

    edgelist = create_edgelist(
        clustered_reads=clustered_reads,
        unique_reads=unique_reads,
        umia_start=umia_start,
        umia_end=umia_end,
        umib_start=umib_start,
        umib_end=umib_end,
        marker=marker,
        sequence=sequence,
    )
    tmp_file = write_tmp_feather_file(edgelist)
    logger.debug(
        "Saved temporary edge list for %s (%s) with %i rows in %s",
        marker,
        sequence,
        len(edgelist),
        tmp_file,
    )
    return tmp_file, input_reads_count



"""Helper classes to define the data input for the qc report.

Copyright © 2022 Pixelgen Technologies AB.
"""

import dataclasses
from typing import List, Optional, TypedDict, TypeVar

from pixelator.mpx.report.models.commands import CommandInfo

CommandOptionValue = TypeVar("CommandOptionValue", str, int, float, bool)


@dataclasses.dataclass
class SampleInfo:
    """Dataclass for general sample information.

    The values defined here are used in the dynamic qc report to populate the
    "sample information" panel.
    """

    #: Version of pixelator used to generate the metrics
    pixelator_version: str

    #: Date the report was generated
    generation_date: str

    #: Identifier of the sample
    sample_id: str

    #: Short description of the sample
    sample_description: str

    #: Version of the pixel
    pixel_version: str

    #: Name of the pixelator panel file
    panel_name: Optional[str]

    #: Version of the pixelator panel file
    panel_version: Optional[str]

    #: List of commands and their parameters
    parameters: List[CommandInfo]


# --- START AUTOGENERATED: metrics ---
# using metrics.json from pixelator-qc-report (0.4.0)
class Metrics(TypedDict, total=False):
    """A typed dictionary containing all metrics that can be used in the qc report.

    Values defined here will be used in the dynamic qc report to populate
    the different data tables. Missing values will be left blank.

    Note: This class is autogenerated from the metrics.json file (0.4.0)
    """

    #: Number of library reads that contain an Antibody DNA-barcode from a Pixelgen
    #: Technologies panel.
    antibody_reads: int

    #: Number of Antibody library reads that contain a recognized Antibody DNA-
    #: barcode, a valid UMI, valid PBS adapters and valid UPI Molecular Pixelation
    #: identifiers.
    antibody_reads_usable_per_cell: int

    #: Number of Antibody library reads that were filtered after annotation and tau-
    #: metric outlier detection.
    antibody_reads_in_outliers: int

    #: Number of Antibody library reads that contain an unrecognizable antibody
    #: barcode.
    unrecognized_antibodies: int

    #: Fraction of library reads that contain an Antibody DNA-barcode from a Pixelgen
    #: Technologies panel.
    fraction_antibody_reads: float

    #: Fraction of Antibody library reads that contain a recognized Antibody DNA-
    #: barcode, a valid UMI, valid PBS adapters and valid UPI Molecular Pixelation
    #: identifiers.
    fraction_antibody_reads_usable_per_cell: float

    #: Fraction of Antibody library reads that were filtered after annotation and
    #: tau-metric outlier detection.
    fraction_antibody_reads_in_outliers: float

    #: Fraction of Antibody library reads that contain an unrecognizable antibody
    #: barcode.
    fraction_unrecognized_antibodies: float

    #: Fraction of Antibody library reads in Cells with a recognized Antibody DNA-
    #: barcode, a valid UMI, valid PBS adapters and valid UPI Molecular Pixelation
    #: identifiers.
    fraction_antibody_reads_in_cells: float

    #: The number of cells annotated by the pipeline.
    number_of_cells: int

    #: Average reads per cell after performing the filtering stages of Pixelator to
    #: retrieve only confidently decoded MPX reads.
    average_reads_usable_per_cell: int

    #: Average reads per cell after Illumina Passed Filter step, removal of phiX as
    #: well as accounting for the actual number of cells.
    average_reads_per_cell: int

    #: The average number of molecules associated to antibodies on annotated cells.
    average_antibody_molecules_per_cell: float

    #: The average number of UMIs per A-pixel.
    average_umis_per_upia: int

    #: The average number of A-pixels per cell identified.
    average_upias_per_cell: int

    #: The percentage of sequenced reads usable in cells. A usable read contains a
    #: recognized Antibody DNA-barcode, a valid UMI, valid PBS adapters and valid UPI
    #: Molecular Pixelation identifiers.
    fraction_reads_in_cells: float

    #: The percentage of UMIs in non-cell components.
    fraction_discarded_umis: float

    #: The total number of antibodies with at least {min-count} reads from molecules
    #: (UMI+UPI) in any cell.
    total_unique_antibodies_detected: int

    #: The fraction of cells that are outliers as determined by the [tau-metric
    #: outlier
    #: detection](https://software.pixelgen.com/pixelator/algorithms#antibody-count-
    #: distribution-outliers).
    fraction_outlier_cells: float

    #: Total number of read pairs that were input to this library.
    number_of_reads: int

    #: Total number of read pairs ignored by the pipeline that do not satisfy the
    #: minimum length requirements.
    number_of_short_reads_skipped: int

    #: Fraction of reads found with valid Molecular Pixelation binding adapter.
    fraction_valid_pbs: float

    #: Fraction of reads with valid molecules that passed filters and were used to
    #: build graph components.
    fraction_valid_umis: float

    #: The proportion of reads from already found molecules (UMI+UPI). This metric
    #: will be a function of library complexity and sequencing depth. More
    #: specifically, this is the fraction of duplicated reads from the same Molecular
    #: Pixelation amplicon.
    sequencing_saturation: float

    #: The average number of reads per molecule is a measure of sequencing depth in
    #: the sample and hints if the underlying molecule distribution has been well
    #: characterized.
    average_reads_per_molecule: float

    #: Fraction of Antibody Barcode bases with Q-score ≥ 30, excluding very low
    #: quality/no-call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_antibody_barcode: float

    #: Fraction of UMI bases with Q-score ≥ 30, excluding very low quality/no-call (Q
    #: ≤ 2) bases from the denominator.
    fraction_q30_bases_in_umi: float

    #: Fraction of UPIA bases with Q-score ≥ 30, excluding very low quality/no-call
    #: (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_upia: float

    #: Fraction of UPIB bases with Q-score ≥ 30, excluding very low quality/no-call
    #: (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_upib: float

    #: Fraction of PBS1 bases with Q-score ≥ 30, excluding very low quality/no-call
    #: (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_pbs1: float

    #: Fraction of PBS2 bases with Q-score ≥ 30, excluding very low quality/no-call
    #: (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_pbs2: float

    #: Fraction of R1 and R2 bases with Q-score ≥ 30, excluding very low quality/no-
    #: call (Q ≤ 2) bases from the denominator.
    fraction_q30_bases_in_read: float


# --- END AUTOGENERATED: metrics ---


class InfoAndMetrics(TypedDict):
    """Aggregated info and metrics."""

    info: SampleInfo
    metrics: Metrics


@dataclasses.dataclass
class QCReportData:
    """Collection of data inputs for dynamic figures in the qc report.

    All attributes are strings containing csv or json data ready to be
    injected into the dynamic report.
    """

    #: CSV formatted string containing a DataFrame for the
    #: "component_size_distribution" figure
    #: component_size, rank, selected, markers
    #: 18463, 1, True, 32
    #: 18290, 2, True, 39
    ranked_component_size: Optional[str]

    #: CSV formatted string containing a DataFrame used for plotting embeddings
    #: with various colors,
    #: eg (cell type, leiden cluster, umis)
    #: Following columns are required:
    #:  component,umap1,umap2,umis,cluster,cluster_cell_class
    #: For example:
    #: ```
    #: component,umap1,umap2,cluster,cluster_cell_class,reads,molecules,mean_upia_degree,mean_umi_per_upia
    #: 0,11.891996,12.216717,4793,2,other
    #: 1,11.02211,12.486254,6708,2,other
    #: 2,7.8869944,7.5836325,5031,5,predict_B-cells
    #: 3,0.66538113,4.00765,3446,9,predict_CD4+_T-cells
    #: 4,13.21712,5.2930307,6704,1,predict_B-cells
    #: ```
    component_data: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the antibodies
    #: per cell figure
    #: Following columns are required: antibodies,reads
    #: For example:
    #: ```
    #: reads,antibodies
    #: 0,0
    #: 5000,5
    #: 5450,10
    #: 10900,15
    #: ```
    antibodies_per_cell: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the sequencing
    #: saturation figure.
    #: Following columns are required: reads,saturation
    #: For example:
    #: ```
    #: reads,saturation
    #: 0,0
    #: 5000,0.1609
    #: 5450,0.1736057
    #: ```
    sequencing_saturation: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the antibodies
    #: per cell figure.
    #: Following columns are required: antibody,percentage
    #: For example:
    #: ```
    #: antibody,percentage
    #: CD274,0.0016104640666519112
    #: CD44,0.04755881148351282
    #: CD56,0.020829214318265954
    #: ```
    antibody_percentages: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the antibody
    #: counts violin plots.
    #: It consists of the component column followed by a list of antibody columns.
    #: For example:
    #: ```
    #: component,CD274,CD44,CD56
    #: PXLCMP0000000,9.0,76.0,63.0
    #: PXLCMP0000002,3.0,90.0,166.0
    #: PXLCMP0000003,1.0,44.0,113.0
    #: ```
    antibody_counts: Optional[str]

    #: CSV formatted string containing a DataFrame used to generate the reads per UMI
    #: histogram.
    #: It consists of a frequency table with absolute and relative counts for the
    #: "count" field of the edgelist
    #: ```
    #: reads_per_molecule,count,frequency
    #: 2, 667338, 0.143356
    #: 3, 675999, 0.145216
    #: 4, 645161, 0.138592
    #: 5, 580468, 0.124695
    #: 6, 491899, 0.105668
    #: 7, 397472, 0.085384
    #: 8, 311331, 0.066879
    #: 9, 236724, 0.050852
    #: 10, 176052, 0.037819
    #: 11, 130251, 0.02798
    #: ```
    reads_per_molecule_frequency: Optional[str]



"""Module containing classes and functions related to the different assay/designs used by pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

import enum
import json
from functools import cache
from typing import Any, List, Mapping, Optional, Set, Tuple

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import pydantic
from pydantic import BaseModel

from pixelator.common.config.utils import load_yaml_file
from pixelator.common.types import PathType


class RegionType(str, enum.Enum):
    """Enum class representing different sequence types."""

    UMI = "umi"
    UPI = "upi"
    PBS = "pbs"
    BARCODE = "barcode"
    AMPLICON = "amplicon"


class SequenceType(str, enum.Enum):
    """Enum class representing different sequence input types.

    JOINED: The sequence given must be empty.
            The sequence will be generated by joining the sequences of the
            sub-regions in the order specified in the assay specification.

    FIXED: The sequence of the regions is fixed and given.
    RANDOM: The sequence of the regions is random
    """

    JOINED = "joined"
    FIXED = "fixed"
    RANDOM = "random"


DNA_CHARS = {"A", "C", "G", "T"}


class RegionModel(BaseModel):
    """Validation model for region configuration."""

    region_id: str
    region_type: RegionType
    name: str
    sequence_type: SequenceType
    sequence: Optional[str] = None
    min_len: Optional[int] = None
    max_len: Optional[int] = None
    data: Optional[Mapping[str, Any]] = {}
    regions: Optional[List["RegionModel"]] = []

    @pydantic.model_validator(mode="after")
    def check_valid_dna_string(self) -> Self:
        """Validate DNA strings.

        :return: the validated model
        :raises ValueError: if the string contains non DNA characters
        """
        if self.region_type is RegionType.PBS:
            is_valid = True
            if self.sequence:
                is_valid = all(character in DNA_CHARS for character in self.sequence)
            if not is_valid:
                raise ValueError("Not a valid DNA string: found non DNA characters")

        return self


class AssayModel(BaseModel):
    """Validation model for assay configuration.

    :ivar name: Name of the assay
    :ivar assay_spec: List of assay regions
    """

    # TODO: Add more metadata fields
    name: str
    assay_spec: List[RegionModel]


class Region:
    """Class representing a region in an assay.

    :ivar region_id: unique ID of the region
    :ivar region_type: type of the region
    :ivar name: name of the region
    :ivar sequence_type: type of the sequence
    :ivar sequence: string representation of the region
    :ivar min_len: minimum length of the region
    :ivar max_len: maximum length of the region
    :ivar regions: list of sub-regions
    :ivar data: additional user-defined region properties
    """

    def __init__(
        self,
        region_id: str,
        region_type: RegionType,
        name: str,
        sequence_type: SequenceType,
        sequence: str = "",
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
        regions: Optional[List["Region"]] = None,
        data: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Initialize a Region."""
        self.parent_id: str | None = None
        self.region_id = region_id
        self.region_type = region_type
        self.name = name
        self.sequence_type = sequence_type
        self.sequence = sequence
        self.data = data

        self.min_len: int
        self.max_len: int

        if self.sequence_type == SequenceType.FIXED:
            if min_len != max_len or min_len != len(self.sequence):
                raise ValueError("Exact sequence does not match min/max length")

            if min_len is None:
                self.min_len = len(self.sequence)

            if max_len is None:
                self.max_len = len(self.sequence)

        self.min_len = min_len if min_len is not None else 0
        self.max_len = max_len if max_len is not None else 0
        self.regions = regions

        if self.regions:
            map(lambda r: r.set_parent_id(self.region_id), self.regions)

            self.min_len, self.max_len = self.get_len()
            self.sequence = self.get_sequence()

    def set_parent_id(self, parent_id: str) -> None:
        """Set the parent id of this region.

        :param parent_id: parent id to set
        """
        self.parent_id = parent_id

    def get_sequence(self) -> str:
        """Return the sequence representation of this region."""
        s = ""
        if self.regions:
            for r in self.regions:
                s += r.get_sequence()
        else:
            if self.sequence:
                s += self.sequence
            elif self.sequence is None:
                s += "X"
        return s

    def get_len(self) -> Tuple[int, int]:
        """Return the minimum and maximum length of this region."""
        min_l, max_l = 0, 0
        if self.regions:
            for r in self.regions:
                res = r.get_len()
                min_l += res[0]
                max_l += res[1]
        else:
            min_l = self.min_len
            max_l = self.max_len

        return min_l, max_l

    def get_max_length(self):
        """Get the maximum length of this region."""
        return self.get_len()[1]

    def update_attr(self):
        """Update the attributes of the region."""
        if self.regions:
            for idx, r in enumerate(self.regions):
                r.update_attr()

        self.sequence = self.get_sequence()
        self.min_len, self.max_len = self.get_len()
        return

    def __repr__(self) -> str:
        """Return str representation of the instance."""
        d = {
            "region_id": self.region_id,
            "region_type": self.region_type,
            "name": self.name,
            "sequence_type": self.sequence_type,
            "sequence": self.sequence,
            "min_len": self.min_len,
            "max_len": self.max_len,
            "regions": self.regions,
        }
        return f"{d}"

    def to_dict(self):
        """Return a dictionary representation of the region."""
        d = {
            "region_id": self.region_id,
            "region_type": self.region_type,
            "name": self.name,
            "sequence_type": self.sequence_type,
            "sequence": self.sequence,
            "min_len": self.min_len,
            "max_len": self.max_len,
            "regions": [i.to_dict() for i in (self.regions or [])],
        }
        return d

    def get_region_by_id(self, region_id: str) -> Optional["Region"]:
        """Lookup a region by the region_id field."""
        if self.region_id == region_id:
            return self

        if self.regions:
            for r in self.regions:
                if r.region_id == region_id:
                    return r

        return None

    def get_regions_by_type(self, region_type: str) -> List["Region"]:
        """Return all regions with specified region_type."""
        found = []

        if self.region_type == region_type:
            found.append(self)

        if self.regions:
            for r in self.regions:
                found.extend(r.get_regions_by_type(region_type))

        return found

    def get_leaves(self) -> List["Region"]:
        """Return a depth-first list of all leaf regions."""
        leaves = []

        if not self.regions:
            leaves.append(self)
        else:
            for r in self.regions:
                leaves.extend(r.get_leaves())

        return leaves

    def get_leaf_region_types(self) -> Set[RegionType]:
        """Return a set of all leaf region types."""
        leaves = self.get_leaves()
        rtypes = set()

        for r in leaves:
            rtypes.add(r.region_type)

        return rtypes

    def get_subregion_ids(self) -> Set[str]:
        """Return a set of the region_ids of all subregions."""
        ids: Set[str] = set()

        if not self.regions:
            return ids

        for r in self.regions:
            ids.add(r.region_id)
            ids.update(r.get_subregion_ids())

        return ids


class Assay:
    """Class representing a specific assay design.

    An assay contains metadata and a list of (possible nested regions) defining
    the structure of the assay reads.

    :ivar name: Name of the assay
    :ivar assay_spec: List of regions defining the assay structure
    """

    def __init__(self, name: str, assay_spec: Optional[List[Region]] = None):
        """Initialize an assay."""
        self.name = name
        self.assay_spec: List[Region] = assay_spec or []

        self._update_parent_ids()

    @classmethod
    def _validate_region_ids(cls, regions):
        region_ids = set()

        for r in regions:
            if r.region_id in region_ids:
                raise ValueError(f"Duplicate region_id detected: {r.region_id}")

            region_ids.add(r.region_id)

            if r.regions:
                cls._validate_region_ids(r.regions)

    @classmethod
    def _parse_regions(cls, region_data: Optional[List[RegionModel]], level=0):
        # Protected against stack overflow by limiting the number of nested regions
        # 20 nested regions is ridiculously high and should never be reached.
        if level > 20:
            raise RecursionError("Too many nested regions")

        regions: List[Region] = []

        if not region_data:
            return regions

        for r in region_data:
            child_regions = None
            if r.regions:
                child_regions = cls._parse_regions(r.regions, level=level + 1)

            opts = r.dict()
            opts.update({"regions": child_regions})
            region_obj = Region(**opts)

            regions.append(region_obj)

        cls._validate_region_ids(regions)
        return regions

    @classmethod
    def _load_assay_model(cls, assay_data: AssayModel) -> "Assay":
        regions = cls._parse_regions(assay_data.assay_spec)
        assay = Assay(name=assay_data.name, assay_spec=regions)
        return assay

    def _update_parent_ids(self) -> None:
        stack: List[Tuple[Region, Optional[str]]] = [
            (region, None) for region in self.assay_spec
        ]

        while len(stack) > 0:
            region, parent_id = stack.pop()
            if parent_id:
                region.set_parent_id(parent_id)

            if region.regions:
                for r in region.regions:
                    stack.append((r, region.region_id))

    @property
    def region_ids(self) -> Set[str]:
        """Return a set with all regions ids in this assay."""
        ids = set()
        for r in self.assay_spec:
            ids.add(r.region_id)
            ids.update(r.get_subregion_ids())

        return ids

    @classmethod
    def from_yaml(cls, filename: PathType) -> "Assay":
        """Parse an assay from a yaml file.

        :return: an Assay instance loaded from the design config
        :param filename: path to a design config file
        """
        yaml_obj = load_yaml_file(filename)
        checked_obj = AssayModel.model_validate(yaml_obj)
        return cls._load_assay_model(checked_obj)

    def to_json(self):
        """Return a json representation of the assay."""
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=False, indent=4)

    def get_sequence(self) -> str:
        """Return a sequence representing this region."""
        s = []
        for region in self.assay_spec:
            s.append(region.get_sequence())

        return "".join(s)

    def get_region_by_id(self, region_id: str) -> Optional[Region]:
        """Retrieve a region by its id.

        :param region_id: id of the region to retrieve
        :return: region with the given id or None if not found
        """
        for r in self.assay_spec:
            maybe_r = r.get_region_by_id(region_id)
            if maybe_r:
                return maybe_r

        return None

    @cache
    def get_regions_by_type(self, region_type: RegionType) -> List[Region]:
        """Retrieve all regions of a given type.

        :param region_type: region type to retrieve
        :return: list of regions with the given type
        """
        regions = []
        for r in self.assay_spec:
            regions.extend(r.get_regions_by_type(region_type))

        return regions

    def __len__(self) -> int:
        """Return the total length of the assay in bases."""
        return sum([region.get_max_length() for region in self.assay_spec])


def get_position_in_parent(assay: Assay, region_id: str) -> Tuple[int, int]:
    """Return the start and end position of a region relative to its parent region.

    This assumes the amplicon consists of only fixed length regions in the path
    from the start of the amplicon up until the region of interest.

    :param assay: assay design
    :param region_id: region id of the amplicon
    :return: tuple with start and end position of the region in the parent region
    :raises ValueError: if the region_id is not found in the assay
    :raises ValueError: if the region has no parent
    :raises ValueError: if the parent_region is not found in the assay
    """
    region = assay.get_region_by_id(region_id)
    if region is None:
        raise ValueError(f"Assay does not contain a region with id {region_id}")

    parent_id = region.parent_id
    if parent_id is None:
        raise ValueError(f"Region with id {region_id} has no parent")

    parent_region = assay.get_region_by_id(parent_id)
    if parent_region is None:
        raise ValueError(f"Assay does not contain a region with id {region.parent_id}")

    leaves = parent_region.get_leaves()

    min_start = 0

    for leave in leaves:
        if leave.region_id == region_id:
            break

        min_start += leave.min_len

    min_stop = min_start + region.min_len

    return min_start, min_stop



"""PNA Pixel File io.

This module handles writing and reading duckdb based pxl files.

Duckdb is an embedded database that supports storing multiple
tables in one portable file.

Pxl files should contain an edgelist table that contains
the edgelist. Futhermore it should contain a number of tables
that encode the adata object. All these table should be prefixed
with `__adata__`, e.g. the adata.obs table should be named
`__adata__obs`.

In addition to the adata tables, and the edgelist, the database
can contain a metadata table, a layouts table, and a proximity
data table.

A full pxl file has (at least) the following tables:

.. code-block:: none

    ┌─────────────────────┐
    │ name                │
    │ ---                 │
    │ str                 │
    ╞═════════════════════╡
    │ __adata__X          │
    │ __adata__obs        │
    │ __adata__var        │
    │ edgelist            │
    │ layouts             │
    │ metadata            │
    │ proximity           │
    └─────────────────────┘

Copyright © 2025 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
import logging
import shutil
import tempfile
from functools import cache
from pathlib import Path
from typing import Iterable, Literal, Sized

import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
from anndata import AnnData
from anndata import concat as anndata_concat

from pixelator.mpx.pixeldataset.utils import update_metrics_anndata
from pixelator.pna.utils.utils import normalize_input_to_list

logger = logging.getLogger(__name__)


class PixelFileWriter:
    """Writer class for PXL files.

    This class takes care of writing data to duckdb based pxl files.
    In addition to writing data tables it will should take care of
    creating any indexes that can be used to speed up queries downstream.

    When writing using this it is recommended to use the writer as a
    context manager.

    .. code-block:: python

        from pixelator.pna.pixeldataset.io import PixelFileWriter

        with PixelFileWriter("my_file.pxl") as writer:
            writer.write_edgelist(Path("edgelist.parquet"))

    """

    def __init__(self, path: Path, exits_ok: bool = False):
        """Initialize the PixelFileWriter.

        :param path: The path to the PXL file.
        :param exists_ok: Whether to remove the file if it exists.
        """
        self.path = path
        self.exists_ok = exits_ok
        if self.exists_ok and self.path.exists():
            self.path.unlink()
        self._connection: duckdb.DuckDBPyConnection = None  # type: ignore

    def open(self):
        """Open a connection to the PXL file."""
        self._connection = duckdb.connect(self.path)

    def close(self):
        """Close the connection to the PXL file."""
        self._connection.close()
        self._connection = None

    def get_connection(self) -> duckdb.DuckDBPyConnection:
        """Get the connection to the PXL file.

        :return: The duckdb connection.
        """
        if self._connection is None:
            raise ValueError("Connection is not open.")
        return self._connection

    def __enter__(self):
        """Open the writer as a context manager."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close the writer context manager."""
        self.close()

    def _write_parquet_file_to_table(
        self, table_name, edgelist_file: Path | list[Path]
    ):
        self._connection.sql(
            f"""
            CREATE OR REPLACE TABLE {table_name} AS
            SELECT * FROM read_parquet($parquet_file);
            """,
            params={
                "parquet_file": [str(f) for f in edgelist_file]
                if isinstance(edgelist_file, list)
                else str(edgelist_file)
            },
        )

    def write_edgelist(self, edgelist: Path | pl.DataFrame) -> None:
        """Write the edgelist to the PXL file.

        :param edgelist: The path to the edgelist parquet file.
        """
        if isinstance(edgelist, Path):
            self._write_parquet_file_to_table("edgelist", edgelist)
            return

        try:
            with tempfile.NamedTemporaryFile(suffix=".parquet") as f:
                file_path = Path(f.name)
                edgelist.write_parquet(file_path)
                self._write_parquet_file_to_table("edgelist", file_path)
        except AttributeError:
            raise ValueError(f"Invalid input type for edgelist: {type(edgelist)}")

    def _clean_existing_adata_tables(self):
        tables = self._connection.sql("SHOW ALL TABLES")
        adata_tables = tables.pl().filter((pl.col("name").str.starts_with("__adata__")))
        for table in adata_tables.iter_rows(named=True):
            self._connection.sql(
                f'DROP TABLE "{table["database"]}"."{table["schema"]}"."{table["name"]}"'
            )

    def write_adata(self, adata: AnnData) -> None:
        """Write the AnnData object to the PXL file.

        :param adata: The AnnData object to write.
        """
        self._clean_existing_adata_tables()

        X = adata.to_df().reset_index(names="index")
        var = adata.var.reset_index(names="index")
        obs = adata.obs.reset_index(names="index")
        uns = adata.uns

        self._connection.sql(
            """
            CREATE TABLE __adata__X AS SELECT * FROM X;
            CREATE TABLE __adata__var AS SELECT * FROM var;
            CREATE TABLE __adata__obs AS SELECT * FROM obs;

            CREATE TABLE __adata__uns (value JSON);
            INSERT INTO __adata__uns VALUES ($uns_data);
            """,
            params={"uns_data": uns},
        )

        for key in adata.obsm:
            obsm_layer = adata.obsm[key].reset_index(names="index")
            self._connection.sql(
                f"""
                CREATE TABLE __adata__obsm_{key} AS SELECT * FROM obsm_layer;
                """,
            )

    def write_metadata(self, metadata: dict) -> None:
        """Write the metadata to the PXL file.

        :param metadata: The metadata dictionary to write.
        """
        self._connection.sql(
            """
            DROP TABLE IF EXISTS metadata;
            CREATE TABLE metadata (value JSON);
            INSERT INTO metadata VALUES ($metadata);
            """,
            params={"metadata": metadata},
        )

    def write_layouts(self, layouts: Path | pl.DataFrame | list[Path]) -> None:
        """Write the layouts to the PXL file.

        :param layouts: The path to the layouts parquet file.
        """
        try:
            if isinstance(layouts, list) or layouts.is_file():  # type: ignore
                self._write_parquet_file_to_table("layouts", layouts)  # type: ignore
                return
        except AttributeError:
            pass

        try:
            with tempfile.NamedTemporaryFile(suffix=".parquet") as f:
                tmp_path = Path(f)
                layouts.write_parquet(tmp_path)  # type: ignore
                self._write_parquet_file_to_table("layouts", tmp_path)
        except AttributeError:
            raise ValueError(f"Invalid input type for layouts: {type(layouts)}")

    def write_proximity(self, proximity: Path | pl.DataFrame) -> None:
        """Write the proximity data to the PXL file.

        :param proximity: The path to the proximity parquet file.
        """
        try:
            if proximity.is_file():  # type: ignore
                self._write_parquet_file_to_table("proximity", proximity)  # type: ignore
                return
        except AttributeError:
            pass

        try:
            with tempfile.NamedTemporaryFile(suffix=".parquet") as f:
                tmp_path = Path(f.name)
                proximity.write_parquet(tmp_path)  # type: ignore
                self._write_parquet_file_to_table("proximity", tmp_path)
        except AttributeError:
            raise ValueError(f"Invalid input type for proximity: {type(proximity)}")


PXL_FILE_MANDATOR_TABLES = [
    "__adata__X",
    "__adata__var",
    "__adata__obs",
    "edgelist",
]
# Should this be a "metadata" be a mandatory table?
PXL_FILE_ADATA_TABLES = ["__adata__X", "__adata__var", "__adata__obs", "__adata__uns"]
PXL_FILE_OTHER_TABLES = ["edgelist", "metadata", "layouts", "proximity"]


class PxlFile:
    """PxlFile represents a a pxl file on disk and provides basic utility methods."""

    def __init__(self, path: Path, sample_name: str | None = None):
        """Initialize the PxlFile."""
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        self.path = path
        self._sample_name = sample_name

    @property
    def sample_name(self) -> str:
        """Return the sample name of the PxlFile."""
        if self._sample_name:
            return self._sample_name
        try:
            return self.metadata()["sample_name"]
        except KeyError:
            raise ValueError(
                f"Could not determine sample name from {self.path} - please provide a sample name."
            )

    def is_pxl_file(self) -> bool:
        """Check if the file is a PXL file."""
        with duckdb.connect(self.path, read_only=True) as con:
            tables = con.sql("SHOW ALL TABLES").to_df()
            return len(
                set(PXL_FILE_MANDATOR_TABLES).intersection(
                    set(tables["name"].to_list())
                )
            ) == len(PXL_FILE_MANDATOR_TABLES)

    def metadata(self) -> dict:
        """Read the metadata from the PXL file."""
        try:
            with duckdb.connect(self.path, read_only=True) as con:
                metadata = con.sql("SELECT * FROM metadata").fetchone()
                return json.loads(metadata[0]) if metadata else {}
        except duckdb.CatalogException:
            return {}

    def __repr__(self) -> str:
        """Return a string representation of the PxlFile."""
        return f"PxlFile({self.path})"

    def __str__(self) -> str:
        """Return a string representation of the PxlFile."""
        return f"{self.path}"

    @staticmethod
    def copy_pxl_file(src: PxlFile, target: Path) -> PxlFile:
        """Copy a PxlFile to a new location.

        :param src: The source PxlFile.
        :param target: The target path.
        :return: The new PxlFile.
        """
        shutil.copy(src.path, target)
        return PxlFile(target)


class PixelDataViewer:
    """PixelDataViewer is used to create virtual views over multiple pxl files.

    The `PixelDataViewer` provides a unified interface over multiple pxl files that
    can be used to query the data in the files. It does this by creating SQL views over all
    the tables in memory. Nota bene that this does not actually materialize any data
    in memory, so the operation it self is very light weight.

    In addition to providing SQL viewes to run queries over, PixelDataViwer class will
    read and concatenate the adata objects from all the files, and cache the resulting
    AnnData object.
    """

    def __init__(
        self,
        sample_name_to_pxl_file_mapping: dict[str, PxlFile],
        adata_join_strategy: Literal["inner", "outer"] = "inner",
    ):
        """Initialize the PixelDataViewer.

        :param sample_name_to_pxl_file_mapping: A dictionary mapping sample names to PxlFile objects.
        :param adata_join_strategy: The strategy to use when joining the adata objects.
        :raises ValueError: If any of the files are not valid PXL files.
        """
        self._db_to_file_mapping = sample_name_to_pxl_file_mapping
        self._normalized_sample_name_mapping = self._map_sample_names_to_db_names(
            sample_name_to_pxl_file_mapping
        )
        # verify all files are PXL files
        invalid_files = [
            pxl_file
            for pxl_file in self._db_to_file_mapping.values()
            if not pxl_file.is_pxl_file()
        ]
        if invalid_files:
            raise ValueError(f"{invalid_files} are not valid PXL files.")

        self._adata_join_strategy = adata_join_strategy
        self._connection: duckdb.DuckDBPyConnection = None  # type: ignore

    def _map_sample_names_to_db_names(
        self, sample_name_to_pxl_file_mapping: dict[str, PxlFile]
    ) -> dict[str, str]:
        """Create a normalized sample name lookup table for working with the duckdb files."""

        def _normalize_sample_name(name: str) -> str:
            # Remove whitespaces and dashes from the sample name
            # and prefix it with "db_", and prefix all names
            # with "db_" to handle when sample names starts with a number.
            return f"db_{name.replace('-', '_').replace(' ', '_')}"

        normalized_names = {
            sample_name: _normalize_sample_name(sample_name)
            for sample_name, _ in sample_name_to_pxl_file_mapping.items()
        }

        if len(normalized_names) != len(sample_name_to_pxl_file_mapping.keys()):
            raise ValueError(
                "Sample names are not unique after normalizing them - please provide sample names that are unique "
                "even when replacing `-` and whitespaces with `_`."
            )

        return normalized_names

    def _get_normalized_name(self, sample_name: str) -> str:
        """Get the normalized name for a sample name."""
        try:
            return self._normalized_sample_name_mapping[sample_name]
        except KeyError:
            raise KeyError(
                f"Sample name {sample_name} not found in the mapping. Available names are: {set(self._normalized_sample_name_mapping.keys())}"
            )

    @staticmethod
    def from_files(pxl_files: Iterable[PxlFile]) -> PixelDataViewer:
        """Create a PixelDataViewer from a list of PxlFile objects.

        This will use the sample names from the PxlFile metadata as sample names.
        """
        return PixelDataViewer(
            {pxl_file.sample_name: pxl_file for pxl_file in pxl_files}
        )

    @staticmethod
    def from_sample_to_file_mappings(
        sample_name_to_pxl_file_mapping: dict[str, PxlFile],
    ) -> PixelDataViewer:
        """Create a PixelDataViewer from a dictionary mapping sample names to PxlFile objects."""
        return PixelDataViewer(sample_name_to_pxl_file_mapping)

    @property
    def sample_to_file_mappings(self) -> dict[str, Path]:
        """Return a dictionary mapping sample names to PxlFile objects."""
        return {
            sample_name: file_.path
            for sample_name, file_ in self._db_to_file_mapping.items()
        }

    def __enter__(self) -> duckdb.DuckDBPyConnection:
        """Open the connection to the PXL files.

        Attach all the files to the connection and create views over the tables.
        :return: The connection object.
        """
        self._connection = duckdb.connect(":memory:")
        self._attach_to_files(self._connection)
        for table in PXL_FILE_OTHER_TABLES:
            self._simple_union_table_view(self._connection, table, table)
        return self._connection

    def __exit__(self, exc_type, exc_value, traceback):
        """Close the connection to the PXL files."""
        self._connection.close()

    def sample_names(self) -> list[str]:
        """Return the list of sample names known to the view."""
        return list(self._db_to_file_mapping.keys())

    def read_adata_from_sample(self, sample: str) -> AnnData:
        """Read the AnnData object from the PXL file.

        :return: The AnnData object.
        """
        with self as connection:
            X = connection.sql(
                f"SELECT * FROM {self._get_normalized_name(sample)}.__adata__X"
            ).to_df()
            var = connection.sql(
                f"SELECT * FROM {self._get_normalized_name(sample)}.__adata__var"
            ).to_df()
            obs = connection.sql(
                f"SELECT * FROM {self._get_normalized_name(sample)}.__adata__obs"
            ).to_df()

            maybe_uns = connection.sql(
                f"select * from {self._get_normalized_name(sample)}.__adata__uns"
            ).fetchone()
            uns = json.loads(maybe_uns[0]) if maybe_uns else None

            tables = connection.sql("SHOW ALL TABLES")

            obsm_tables = (
                tables.pl()
                .filter(
                    (pl.col("name").str.starts_with("__adata__obsm"))
                    & (pl.col("database") == self._get_normalized_name(sample))
                )
                .select(
                    pl.concat_str(
                        [pl.col("database"), pl.col("schema"), pl.col("name")],
                        separator=".",
                    ).alias("name")
                )
                .get_column("name")
                .to_list()
            )

            obsm = {
                table.split("__adata__obsm_")[1]: (
                    connection.sql(f"SELECT * FROM {table}")
                    .to_df()
                    .set_index("index")
                    .rename_axis(index={"index": "component"})
                )
                for table in obsm_tables
            }

            adata = AnnData(
                X=X.set_index("index").rename_axis(index={"index": "component"}),
                var=var.set_index("index").rename_axis(index={"index": "marker_id"}),
                obs=obs.set_index("index").rename_axis(index={"index": "component"}),
                uns=uns,
                obsm=obsm,
            )
            return adata

    @cache
    def read_adata(self) -> AnnData:
        """Read the AnnData object from the PXL file.

        The result will be cached to avoid reading and concatenating
        the anndata multiple times.

        :return: The AnnData object
        """

        def underlying_data():
            for sample_name, _ in self._db_to_file_mapping.items():
                adata = self.read_adata_from_sample(sample_name)
                adata.obs["sample"] = sample_name
                yield adata

        adatas = list(underlying_data())
        concatenated = anndata_concat(adatas, join=self._adata_join_strategy)
        concatenated.var = adatas[0].var
        concatenated.uns = adatas[0].uns
        update_metrics_anndata(concatenated, inplace=True)
        return concatenated

    def _attach_to_files(self, connection: duckdb.DuckDBPyConnection):
        for name, path in self._db_to_file_mapping.items():
            query = f"ATTACH DATABASE '{path}' AS {self._get_normalized_name(name)} (READ_ONLY);"
            connection.execute(query)

    def _simple_union_table_view(
        self, connection: duckdb.DuckDBPyConnection, table_name: str, view_name: str
    ):
        """Create a view that unions given table from all the underlying samples."""
        # This will create a view tha unions the tables from all the samples,
        # adding a new column that identifies the sample.

        try:
            view_query = [f"""CREATE VIEW {view_name} AS"""]

            table_queries = []
            for sample_name, _ in self._db_to_file_mapping.items():
                table_queries.append(
                    f"SELECT *, '{sample_name}' as 'sample' FROM {self._get_normalized_name(sample_name)}.{table_name}"
                )
            union_query = "\n UNION ALL \n".join(table_queries)

            view_query.append(union_query)
            view_query_str = "\n".join(view_query)

            connection.execute(view_query_str)
        except duckdb.CatalogException as e:
            if table_name in PXL_FILE_MANDATOR_TABLES:
                raise ValueError(
                    f"Mandatory table {table_name} is missing from {self._db_to_file_mapping[sample_name]}- are you sure this is a pxl file?"
                )
            # note that we are ignoring the exception here, since it is expected
            # that some tables may not be present in all files.
            pass


class PixelDataQuerier:
    """Class to read data from a PXL view - this view an represent one or more underlying files on disk."""

    def __init__(self, view: PixelDataViewer):
        """Initialize the PixelDataQuerier."""
        self.view = view

    def read_adata(self) -> AnnData:
        """Read the AnnData object from the PXL file.

        :return: The AnnData object.
        """
        return self.view.read_adata()

    def read_all_component_names(self) -> set[str]:
        """Read all component names from the PXL file.

        :return: A set of component names.
        """
        return set(self.view.read_adata().obs.index.to_list())

    def read_all_marker_names(self) -> set[str]:
        """Read all marker names from the PXL file.

        :return: A set of component names.
        """
        return set(self.view.read_adata().var.index.to_list())

    def _optimized_component_where_condition(self, components: Sized | None) -> str:
        """Create an optimized where clause depending on the number of components.

        Since duckdb does not support predicate pushdown for IN clauses, we need to
        optimize the query by using an equal statement to speed things up when
        selecting a single component.

        This may change in later version of duckdb, see conversation here:
        https://discord.com/channels/909674491309850675/1032659480539824208/1336979122512986144
        """
        if not components:
            return "TRUE"
        if len(components) == 1:
            return f"component = $components"
        return f"component IN $components"

    def read_edgelist(
        self,
        components: Iterable[str] | str | None = None,
        as_pandas=False,
    ) -> pl.DataFrame | pd.DataFrame:
        """Read the edgelist from the PXL file.

        :param components: The components to filter by.
        :return: A DataFrame containing the edgelist.
        """
        with self.view as connection:
            components = normalize_input_to_list(components)
            query = f"""SELECT * FROM edgelist
                        WHERE {self._optimized_component_where_condition(components)}
                    """
            params = {}
            if components is not None:
                params["components"] = (
                    components if len(components) > 1 else components[0]
                )
            if as_pandas:
                return connection.sql(query, params=params).df()
            return connection.sql(query, params=params).pl()

    def read_edgelist_len(
        self,
        components: Iterable[str] | str | None = None,
    ) -> int:
        """Get the length of the edgelist.

        :param components: The components to filter by.
        :return: The length of the edgelist.
        """
        components = normalize_input_to_list(components)

        query = f"""SELECT COUNT(*) FROM edgelist
                    WHERE {self._optimized_component_where_condition(components)}
        """
        params = {}
        if components is not None:
            params["components"] = components if len(components) > 1 else components[0]
        with self.view as connection:
            return connection.sql(query, params=params).execute().fetchone()[0]  # type: ignore

    def read_edgelist_stream(
        self,
        components: Iterable[str] | str | None = None,
        batch_size: int = 1_000_000,
    ) -> Iterable[pa.RecordBatch]:
        """Stream the edgelist from the PXL file.

        :param components: The components to filter by.
        :param batch_size: The batch size for streaming.
        :return: An iterable of RecordBatches.
        """
        components = normalize_input_to_list(components)

        query = f"""SELECT * FROM edgelist
                    WHERE {self._optimized_component_where_condition(components)}
                """
        params = {}
        if components is not None:
            params["components"] = components if len(components) > 1 else components[0]

        with self.view as connection:
            result = connection.sql(query, params=params)
            reader = result.fetch_arrow_reader(batch_size=batch_size)
            for batch in reader:
                yield batch

    def read_metadata(self) -> dict:
        """Read the metadata from the PXL file.

        :return: The metadata dictionary.
        """
        with self.view as connection:
            maybe_metadata = list(
                map(
                    lambda x: json.loads(x[0]),
                    connection.sql("SELECT * FROM metadata").fetchall(),
                )
            )
            if not maybe_metadata:
                return {}

            metadata = {}
            for metadata_dict in maybe_metadata:
                metadata[metadata_dict["sample_name"]] = metadata_dict
        return metadata

    def read_layouts(
        self,
        components: str | Iterable[str] | None = None,
        add_marker_counts: bool = False,
    ) -> pl.DataFrame:
        """Read the layouts from the PXL file.

        :param components: The components to filter by.
        :param add_marker_counts: Whether to add marker counts.
        :return: A DataFrame containing the layouts.
        """

        def _pivot_marker_table(df):
            return (
                df.select(pl.col("*"), val=pl.lit(1))
                .pivot(
                    on="marker",
                    index=None,
                    values="val",
                    aggregate_function=pl.len().cast(pl.UInt8),
                )
                .fill_null(0)
            )

        components = normalize_input_to_list(components)

        if add_marker_counts:
            query = f"""
                WITH filtered_edgelist AS (
                    SELECT umi1 as umi, marker_1 as marker
                    FROM edgelist
                    WHERE {self._optimized_component_where_condition(components)}

                    UNION

                    SELECT umi2 as umi, marker_2 as marker
                    FROM edgelist
                    WHERE {self._optimized_component_where_condition(components)}
                )
                SELECT *
                FROM layouts
                LEFT JOIN filtered_edgelist as umi_and_markers
                ON layouts.index = umi_and_markers.umi
                WHERE {self._optimized_component_where_condition(components)}
            """
        else:
            query = f"""SELECT * FROM layouts
                        WHERE {self._optimized_component_where_condition(components)}
                    """

        try:
            params = {}
            if components is not None:
                params["components"] = (
                    components if len(components) > 1 else components[0]
                )
            with self.view as connection:
                result = connection.sql(query, params=params).pl()
                result = _pivot_marker_table(result) if add_marker_counts else result
                return result.drop(["umi", "marker"], strict=False)
        except duckdb.CatalogException:
            return pl.DataFrame()

    def read_layouts_len(
        self,
        components: str | Iterable[str] | None = None,
    ) -> int:
        """Get the length of the layouts.

        :param components: The components to filter by.
        :return: The length of the layouts.
        """
        components = normalize_input_to_list(components)  # type: ignore
        query = f"""SELECT COUNT(*)
                    FROM layouts
                    WHERE {self._optimized_component_where_condition(components)}
                """
        try:
            params = {}
            if components is not None:
                params["components"] = (
                    components if len(components) > 1 else components[0]  # type: ignore
                )
            with self.view as connection:
                return connection.sql(query, params=params).execute().fetchone()[0]  # type: ignore
        except duckdb.CatalogException:
            return 0

    def read_proximity(
        self,
        components: str | Iterable[str] | None = None,
        markers: str | Iterable[str] | None = None,
    ) -> pl.DataFrame:
        """Read the proximity data from the PXL file.

        :param components: The components to filter by.
        :param markers: The markers to filter by.
        :return: A DataFrame containing the proximity data.
        """
        components = normalize_input_to_list(components)  # type: ignore
        markers = normalize_input_to_list(markers)  # type: ignore
        try:
            query = f"""SELECT * FROM proximity
                         WHERE  {self._optimized_component_where_condition(components)} AND
                                {"(marker_1 IN $markers AND marker_2 IN $markers)" if markers else "TRUE"};
                    """
            params = {}
            if components is not None:
                params["components"] = (
                    components if len(components) > 1 else components[0]  # type: ignore
                )
            if markers is not None:
                params["markers"] = markers
            with self.view as connection:
                return connection.sql(query, params=params).pl()
        except duckdb.CatalogException:
            return pl.DataFrame()

    def read_proximity_len(
        self,
        components: str | Iterable[str] | None = None,
        markers: str | Iterable[str] | None = None,
    ) -> int:
        """Get the length of the proximity data.

        :param components: The components to filter by.
        :param markers: The markers to filter by.
        :return: The length of the proximity data.
        """
        components = normalize_input_to_list(components)  # type: ignore
        markers = normalize_input_to_list(markers)  # type: ignore
        try:
            query = f"""SELECT COUNT(*) FROM proximity
                        WHERE {self._optimized_component_where_condition(components)} AND
                              {"(marker_1 IN $markers AND marker_2 IN $markers)" if markers else "TRUE"};
                    """
            params = {}
            if components is not None:
                params["components"] = components
            if markers is not None:
                params["markers"] = markers
            with self.view as connection:
                return connection.sql(query, params=params).execute().fetchone()[0]  # type: ignore
        except duckdb.CatalogException:
            return 0


class InplacePixelDataFilterer:
    """Class to filter a PXL file in place.

    This is mostly useful for testing purposes, when one wants to strip
    componentens from a PXL file to make it smaller and faster to work with.
    """

    def __init__(self, pxl_file: PxlFile):
        """Initialize the InplacePixelDataFilterer."""
        self.pxl_file = pxl_file

    def _update_metadata(
        self, connection: duckdb.DuckDBPyConnection, metadata: dict
    ) -> None:
        connection.sql(
            """
            DROP TABLE IF EXISTS metadata;
            CREATE TABLE metadata (value JSON);
            INSERT INTO metadata VALUES ($metadata);
            """,
            params={"metadata": metadata},
        )

    def _filter_edgelist(
        self, connection: duckdb.DuckDBPyConnection, components: list[str]
    ) -> None:
        query = f"""
            DELETE FROM edgelist
            WHERE component NOT IN $components
        """
        connection.sql(query, params={"components": components})

    def _filter_proximity(
        self, connection: duckdb.DuckDBPyConnection, components: list[str]
    ) -> None:
        try:
            query = f"""
                DELETE FROM proximity
                WHERE component NOT IN $components
            """
            connection.sql(query, params={"components": components})
        except duckdb.CatalogException:
            pass

    def _filter_layouts(
        self, connection: duckdb.DuckDBPyConnection, components: list[str]
    ) -> None:
        try:
            query = f"""
                DELETE FROM layouts
                WHERE component NOT IN $components
            """
            connection.sql(query, params={"components": components})
        except duckdb.CatalogException:
            pass

    def _filter_adata(self, pxl_file: PxlFile, components: list[str]) -> None:
        adata = PixelDataQuerier(PixelDataViewer.from_files([pxl_file])).read_adata()
        adata = adata[adata.obs.index.isin(components)]
        with PixelFileWriter(pxl_file.path) as writer:
            writer.write_adata(adata)

    def filter_components(
        self, components: set[str] | list[str], metadata: dict | None = None
    ) -> None:
        """Filter the PXL file by components, only keeping the provided components.

        Note that if you provide metadata it will overwrite the existing metadata.
        If your do not provide metadata, the existing metadata will be kept.

        :param components: The components to keep.
        :param metadata: The metadata to write to the PXL
        """
        # I there are not components provided, do nothing.
        if not components:
            raise ValueError("You must provided at least one component to filter.")
        components_as_list: list[str] = normalize_input_to_list(components)  # type: ignore

        with duckdb.connect(self.pxl_file.path) as connection:
            self._filter_edgelist(connection, components_as_list)
            self._filter_proximity(connection, components_as_list)
            self._filter_layouts(connection, components_as_list)
            if metadata:
                self._update_metadata(connection, metadata)

        self._filter_adata(self.pxl_file, components_as_list)


def copy_databases(src_db: Path, target_db: Path) -> None:
    """Copy the contents of one PXL file to another.

    This is a trick that can be used to reclaim disk-space. See:
    https://duckdb.org/docs/stable/operations_manual/footprint_of_duckdb/reclaiming_space.html

    :param src_db: The source PXL file.
    :param target_db: The target PXL file.
    """
    query = f"""
    ATTACH '{str(src_db)}' AS src (READ_ONLY);
    ATTACH '{str(target_db)}' AS target;
    COPY FROM DATABASE src TO target;
    """

    with duckdb.connect() as connection:
        connection.execute(query)



"""Correction and demultiplexing of marker barcodes in a FASTQ file.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
import sys
import typing
from pathlib import Path, PurePath
from typing import Literal

import duckdb as dd
import pyarrow as pa
import pyarrow.parquet
from cutadapt.files import FileOpener, InputPaths, OutputFiles
from cutadapt.pipeline import SingleEndPipeline
from cutadapt.steps import SingleEndSink
from cutadapt.utils import DummyProgress, Progress

from pixelator.common.exceptions import PixelatorBaseException
from pixelator.common.utils import get_sample_name
from pixelator.pna.config import PNAAntibodyPanel, PNAAssay
from pixelator.pna.demux.barcode_demuxer import (
    BarcodeDemuxer,
    IndependentBarcodeDemuxer,
    PairedBarcodeDemuxer,
    create_barcode_group_to_batch_mapping,
    independent_marker_groups_mapping,
)
from pixelator.pna.demux.barcode_identifier import BarcodeIdentifier
from pixelator.pna.demux.pipeline import (
    DemuxFilenamePolicy,
    DemuxPipeline,
    IndependentMarkersFilenamePolicy,
    ParallelDemuxPipelineRunner,
    PartsFilenamePolicy,
)
from pixelator.pna.demux.report import BarcodeCorrectionStatistics
from pixelator.pna.read_processing.runners import ParallelPipelineRunner
from pixelator.pna.utils import clean_suffixes, init_duckdb_conn

logger = logging.getLogger("demux")

import multiprocessing as mp


def correct_marker_barcodes(
    input: Path,
    assay: PNAAssay,
    panel: PNAAntibodyPanel,
    output: Path,
    save_failed: bool = True,
    mismatches: int = 1,
    threads: int = -1,
) -> tuple[BarcodeCorrectionStatistics, Path, Path | None]:
    """Correct the marker barcodes in a FASTQ file.

    Errors in the sequence barcodes within `mismatches` distance are
    corrected to the list of known markers in the panel.
    The original reads are left unmodified but the name of the
    marker is added to the header.

    :param input: The path to the input FASTQ file.
    :param assay: The assay design.
    :param panel: The antibody panel.
    :param output: The output directory.
    :param save_failed: Save failed reads to a separate file.
    :param mismatches: The number of mismatches allowed when correcting towards known markers.
    :param threads: The number of threads to use for processing. By default all available cores are used.
    """
    threads = threads if threads > 0 else mp.cpu_count()

    # Open file handles for input files
    input_files = InputPaths(str(input))

    # Open file handles for output files
    # The compression threads will be capped at 1
    compression_threads = 1
    file_opener = FileOpener(threads=compression_threads)

    logger.info(
        "Opening output file writer with %s compression threads", compression_threads
    )

    # Open file handles for output files
    # When writing from multiple cores the `proxied` argument will take care
    # of the necessary reordering.
    output_files = OutputFiles(
        proxied=True,
        qualities=True,
        file_opener=file_opener,
        interleaved=False,
    )

    # Open file handles for corrected and failed reads
    sample_name = get_sample_name(input)
    corrected_reads_path = Path(output / f"{sample_name}.demux.passed.fq.zst")
    failed_reads_path = None
    if save_failed:
        failed_reads_path = Path(output / f"{sample_name}.demux.failed.fq.zst")

    sink = SingleEndSink(output_files.open_record_writer(corrected_reads_path))

    failed_writer = None
    if save_failed:
        failed_writer = output_files.open_record_writer(failed_reads_path)

    barcodes_id = BarcodeIdentifier(
        assay=assay, panel=panel, mismatches=mismatches, writer=failed_writer
    )

    # Construct the pipeline
    pipeline = SingleEndPipeline(modifiers=[], steps=[barcodes_id, sink])

    # Progress bar for the pipeline
    if sys.stderr.isatty():
        progress = Progress()
    else:
        progress = DummyProgress()

    n_workers = max(1, threads - compression_threads)
    logger.info("Correcting barcodes using %s worker threads", n_workers)

    # Run the pipeline on a parallel runner
    runner = ParallelPipelineRunner(
        inpaths=input_files,
        n_workers=n_workers,
        statistics_class=BarcodeCorrectionStatistics,
    )

    with runner as r:
        stats = r.run(pipeline, progress, output_files)
        output_files.close()

    return (stats, corrected_reads_path, failed_reads_path)


def demux_barcode_groups(
    corrected_reads: Path,
    assay: PNAAssay,
    panel: PNAAntibodyPanel,
    stats: BarcodeCorrectionStatistics,
    output_dir: Path,
    reads_per_chunk: int = 50_000_000,
    max_chunks: int = 8,
    threads: int = -1,
    stategy: Literal["independent", "paired"] = "paired",
):
    """Demux a FASTQ file containing marker information in the header.

    Since there are many combinations of markers, the barcode groups are
    not demultiplexed in individual (pid1, pid1) combinations but in batches.
    Sequence records will be streamed into Arrow IPC files given a target batch size.

    Args:
        corrected_reads (Path): The path to the corrected FASTQ file.
        assay (PNAAssay): The assay design.
        panel (PNAAntibodyPanel): The antibody panel.
        stats (BarcodeCorrectionStatistics):
            The statistics from the barcode correction.
        output_dir (Path):
            The output directory.
        reads_per_chunk (int, optional):
            The target number of molecules in each batch.
        max_chunks (int, optional):
            The maximum number of batches.
        threads (int, optional):
            The number of threads to use for processing. The default of `-1` will use all available cores.
        stategy (Literal["independent", "paired"], optional):
            The demultiplexing strategy to use. Defaults to "paired".

    """
    # Open file handles for input files

    threads = threads if threads > 0 else max(mp.cpu_count(), 1)

    input_files = InputPaths(str(corrected_reads))

    # 1 Reader and 1 Writer thread but these do not require much CPU
    worker_threads = max(threads - 1, 1)
    prefix = clean_suffixes(PurePath(input_files.paths[0])).name
    prefix = prefix.replace(".demux.passed", ".demux")
    filename_policy: DemuxFilenamePolicy

    if stategy == "independent":
        m1_group_map, m2_group_map = independent_marker_groups_mapping(
            stats.pid_pair_counter,
            reads_per_chunk=reads_per_chunk,
            max_chunks=max_chunks,
        )
        barcode_demuxer = typing.cast(
            BarcodeDemuxer,
            IndependentBarcodeDemuxer(
                assay=assay,
                panel=panel,
                marker1_groups=m1_group_map,
                marker2_groups=m2_group_map,
            ),
        )
        filename_policy = typing.cast(
            DemuxFilenamePolicy,
            IndependentMarkersFilenamePolicy(
                prefix=prefix,
                m1_map=m1_group_map,
                m2_map=m2_group_map,
            ),
        )

    else:
        batch_mapping = create_barcode_group_to_batch_mapping(
            stats.pid_pair_counter,
            reads_per_chunk=reads_per_chunk,
            max_chunks=max_chunks,
        )
        barcode_demuxer = typing.cast(
            BarcodeDemuxer,
            PairedBarcodeDemuxer(assay=assay, panel=panel, supergroups=batch_mapping),
        )
        filename_policy = typing.cast(
            DemuxFilenamePolicy, PartsFilenamePolicy(prefix=prefix)
        )

    pipeline = DemuxPipeline(barcode_demuxer)
    runner = ParallelDemuxPipelineRunner(
        inpaths=input_files,
        n_workers=worker_threads,
        output_directory=output_dir,
        filename_policy=filename_policy,
    )
    progress = Progress()

    logger.info("Demuxing with %s worker threads", worker_threads)

    with runner as r:
        r.run(pipeline, progress)

    return output_dir


def finalize_batched_groups(
    input_dir: Path,
    output_dir: Path,
    remove_intermediates: bool = True,
    strategy: Literal["paired", "independent"] = "independent",
    memory: int | None = None,
    threads: int | None = None,
    temp_dir: str | Path | None = None,
):
    """Post-process the demuxed data by sorting and writing to Parquet.

    The demuxed data is written to Arrow IPC file format (v2) in the work directory.
    These are then sorted and written to Parquet format with ZSTD compression.

    The sorting is done on the `marker_1` and `marker_2` columns and written to parquet metadata
    which allows for fast contiguous reads [marker_1, marker_2] groups.

    Params:
        input_dir: the path to the work directory containing the demuxed data
        output_dir: the path to the output directory
        remove_intermediates: whether to remove the intermediate Arrow files after writing to parquet
        strategy: the demultiplexing strategy to use. Can be "paired" or "independent"
        memory: maximum amount of memory to use in bytes
    """
    if strategy == "independent":
        return _finalize_batched_groups_independent(
            input_dir,
            output_dir,
            remove_intermediates=remove_intermediates,
            memory=memory,
            threads=threads,
            temp_dir=temp_dir,
        )
    elif strategy == "paired":
        return _finalize_batched_groups_paired(
            input_dir,
            output_dir,
            remove_intermediates=remove_intermediates,
            memory=memory,
            threads=threads,
            temp_dir=temp_dir,
        )
    else:
        raise ValueError("Unknown strategy")


def _finalize_batched_groups_paired(
    input_dir: Path,
    output_dir,
    remove_intermediates: bool = True,
    memory: int | None = None,
    threads: int | None = None,
    temp_dir: str | Path | None = None,
):
    """Post-process the demuxed data by sorting and writing to Parquet.

    The demuxed data is written as intermediate parquet files in the work directory.
    These are then sorted and deduplicated.

    The sorting is done on the `marker_1` and `marker_2` columns and written to parquet metadata
    which allows for fast contiguous reads of [marker_1, marker_2] groups.

    Params:
        input_dir:
            the path to the work directory containing the demuxed data
        output_dir:
            the path to the output directory where the final parquet files are written
        remove_intermediates:
            Whether to remove the intermediate parquet files after sorting and deduplication
        memory:
            Maximum amount of memory to use. Use None to disable memory limits.

    Returns:
        A list of paths to the Parquet files

    Raises:
        ValueError: If no marker identifier (m1 or m2) is found in the Arrow IPC file name.

    """
    parquet_files = []
    tmp_parquet_files = list(input_dir.glob("*.parquet"))
    conn = init_duckdb_conn(memory_limit=memory, threads=threads, temp_dir=temp_dir)

    for f in tmp_parquet_files:
        output_name = str(clean_suffixes(f).name)
        output_name = output_name.removesuffix(".parquet")

        output_path = Path(output_dir) / f"{output_name}.parquet"
        parquet_files.append(output_path)

        conn.execute(
            f"""
            COPY (
               SELECT *, count(*) as read_count
               FROM read_parquet('{f}')
               GROUP BY ALL
               ORDER BY marker_1, marker_2
           ) TO '{output_path}' (FORMAT parquet, COMPRESSION zstd, COMPRESSION_LEVEL 6);
           """
        )

        if remove_intermediates:
            f.unlink()

    return parquet_files


def _finalize_batched_groups_independent(
    input_dir: Path,
    output_dir: Path,
    remove_intermediates: bool = True,
    memory: int | None = None,
    threads: int | None = None,
    temp_dir: str | Path | None = None,
):
    """Post-process the demuxed data by sorting and writing to Parquet.

    The demuxed data is written to Arrow IPC file format (v2) in the work directory.
    These are then sorted and written to Parquet format with ZSTD compression.

    The sorting is done on the `marker_1` and `marker_2` columns and written to parquet metadata
    which allows for fast contiguous reads [marker_1, marker_2] groups.

    Params:
        work_dir:
            the path to the work directory containing the demuxed data
        remove_intermediates:
             Whether to remove the intermediate parquet files after sorting and deduplication
        memory:
            Maximum amount of memory to use for DuckDB operations. Use None to disable memory limits.
        threads:
            The number of threads to use for DuckDB operations. Use None to let DuckDB decide.
        temp_dir:
            The temporary directory to use for DuckDB operations. If None, DuckDB will decide (defaults to /tmp).

    Returns:
        A list of paths to the Parquet files

    Raises:
        ValueError: If no marker identifier (m1 or m2) is found in the Arrow IPC file name.

    """
    parquet_files = []
    tmp_files = list(input_dir.glob("*.parquet"))

    conn = init_duckdb_conn(memory_limit=memory, threads=threads, temp_dir=temp_dir)

    for f in tmp_files:
        sorting_order: tuple[str, str]

        if ".m1." in str(f):
            sorting_order = ("marker_1", "marker_2")
        elif ".m2." in str(f):
            sorting_order = ("marker_2", "marker_1")
        else:
            raise PixelatorBaseException(
                f"Unrecognised marker suffix. Could not determine sorting order"
            )

        output_name = str(clean_suffixes(f).name)
        output_name = output_name.removesuffix(".parquet")

        output_path = Path(output_dir) / f"{output_name}.parquet"
        parquet_files.append(output_path)

        # Params not supported in ORDER BY clause
        conn.execute(f"""
            COPY (
                SELECT *, count(*) as read_count
                FROM read_parquet('{f}')
                GROUP BY ALL
                ORDER BY {sorting_order[0]}, {sorting_order[1]}
            ) TO '{output_path}' (FORMAT parquet, COMPRESSION zstd, COMPRESSION_LEVEL 6);
        """)

        if remove_intermediates:
            f.unlink()

    return parquet_files



"""Model for report data returned by the single-cell-pna graph stage.

Copyright © 2024 Pixelgen Technologies AB.
"""

from __future__ import annotations

from dataclasses import asdict, dataclass

import pydantic

from pixelator.pna.report.models.base import SampleReport


@dataclass
class GraphStatistics:
    """Graph statistics for a sample."""

    # Raw input from collapse to graph
    molecules_input: int = 0
    reads_input: int = 0

    # Molecules and reads after removing UMI collisions
    molecules_post_umi_collision_removal: int = 0
    reads_post_umi_collision_removal: int = 0

    # Input into to initial graph after read_count filtering
    molecules_post_read_count_filtering: int = 0
    reads_post_read_count_filtering: int = 0

    # Final output after recovery and component size filtering
    molecules_output: int = 0
    reads_output: int = 0

    component_count_pre_recovery: int = 0
    component_count_post_recovery: int = 0

    component_count_pre_component_size_filtering: int = 0
    component_count_post_component_size_filtering: int = 0

    fraction_nodes_in_largest_component_pre_recovery: float = 0.0
    fraction_nodes_in_largest_component_post_recovery: float = 0.0
    component_size_min_filtering_threshold: int | None = 0
    component_size_max_filtering_threshold: int | None = None
    crossing_edges_removed: int = 0
    crossing_edges_removed_initial_stage: int = 0
    max_recursion_depth: int = 0

    node_count_pre_recovery: int = 0
    edge_count_pre_recovery: int = 0

    node_count_post_recovery: int = 0
    edge_count_post_recovery: int = 0

    pre_filtering_component_sizes: dict[int, int] | None = None

    median_reads_per_component: float = 0

    median_markers_per_component: float = 0

    aggregate_count: int = 0
    read_count_in_aggregates: int = 0
    edge_count_in_aggregates: int = 0
    node_count_in_aggregates: int = 0

    def to_dict(self):
        """Convert the object to a dictionary."""
        return asdict(self)


class GraphSampleReport(SampleReport):
    """Model for report data returned by the graph stage."""

    report_type: str = "graph"

    molecules_input: int = pydantic.Field(
        ..., description="The number of molecules in the input edgelist."
    )

    reads_input: int = pydantic.Field(
        ..., description="The number of reads in the input edgelist."
    )

    molecules_post_umi_collision_removal: int = pydantic.Field(
        ..., description="The number of molecules after removing UMI collisions."
    )

    reads_post_umi_collision_removal: int = pydantic.Field(
        ..., description="The number of reads after removing UMI collisions."
    )

    molecules_post_read_count_filtering: int = pydantic.Field(
        ..., description="The number of molecules after read count filtering."
    )

    reads_post_read_count_filtering: int = pydantic.Field(
        ..., description="The number of reads after read count filtering."
    )

    molecules_output: int = pydantic.Field(
        ..., description="The number of molecules in the output graph."
    )

    reads_output: int = pydantic.Field(
        ..., description="The number of reads in the output graph."
    )

    component_count_pre_recovery: int = pydantic.Field(
        ...,
        description="The total number of components in the graph before mega-cluster recovery.",
    )

    component_count_post_recovery: int = pydantic.Field(
        ...,
        description="The total number of components in the graph after mega-cluster recovery.",
    )

    component_count_pre_component_size_filtering: int = pydantic.Field(
        ...,
        description="The total number of components in the graph before filtering for size.",
    )

    component_count_post_component_size_filtering: int = pydantic.Field(
        ...,
        description="The total number of components in the graph after filtering for size.",
    )

    fraction_nodes_in_largest_component_pre_recovery: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
        description="The fraction of all molecules that are located in the largest component before multiplet recovery.",
    )

    fraction_nodes_in_largest_component_post_recovery: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
        description="The fraction of all molecules that are located in the largest component after multiplet recovery.",
    )

    crossing_edges_removed: int = pydantic.Field(
        ...,
        description="The total number of crossing edges removed. Will be 0 if multiplet recovery was not enabled.",
    )

    crossing_edges_removed_initial_stage: int = pydantic.Field(
        ...,
        description="The number of crossing edges removed in the initial step. Will be 0 if multiplet recovery was not enabled.",
    )

    component_size_min_filtering_threshold: int | None = pydantic.Field(
        ..., description="The minimum component size threshold used for filtering."
    )
    component_size_max_filtering_threshold: int | None = pydantic.Field(
        ..., description="The max component size threshold used for filtering."
    )

    node_count_pre_recovery: int = pydantic.Field(
        ..., description="The number of nodes before multiplet recovery."
    )
    edge_count_pre_recovery: int = pydantic.Field(
        ..., description="The number of edges before multiplet recovery."
    )

    node_count_post_recovery: int = pydantic.Field(
        ..., description="The number of nodes after multiplet recovery."
    )
    edge_count_post_recovery: int = pydantic.Field(
        ..., description="The number of edges after multiplet recovery."
    )

    pre_filtering_component_sizes: dict[int, int] = pydantic.Field(
        ..., description="The sizes of all components before filtering."
    )

    median_reads_per_component: float = pydantic.Field(
        ..., description="The median number of reads per cell component."
    )

    median_markers_per_component: float = pydantic.Field(
        ..., description="The median number of markers per cell component."
    )

    aggregate_count: int = pydantic.Field(
        ..., description="The number of aggregates in the graph."
    )

    read_count_in_aggregates: int = pydantic.Field(
        ..., description="The number of reads in aggregates."
    )

    edge_count_in_aggregates: int = pydantic.Field(
        ..., description="The number of edges in aggregates."
    )

    @pydantic.computed_field(  # type: ignore
        description="The fraction of components discarded by filtering.",
        return_type=float,
    )
    @property
    def fraction_of_discarded_components(self) -> float:
        """Calculate the fraction of discarded components."""
        if self.component_count_pre_component_size_filtering == 0:
            return 0.0
        return (
            1
            - self.component_count_post_component_size_filtering
            / self.component_count_pre_component_size_filtering
        )

    @pydantic.computed_field(  # type: ignore
        description="The fraction of components marked as aggregates by the tau_type outlier detection. ",
        return_type=float,
    )  # type: ignore
    @property
    def fraction_of_aggregate_components(self) -> float:
        """Calculate the fraction of aggregate components."""
        if self.component_count_post_component_size_filtering == 0:
            return 0.0
        return self.aggregate_count / self.component_count_post_component_size_filtering

    @pydantic.computed_field(  # type: ignore
        description="The redundancy of edge detection. The saturation is calculated as: `1 - # graph edges / # graph reads`",
        return_type=float,
    )
    @property
    def edge_saturation(self):
        """Return the edge saturation."""
        return 1 - self.edge_count_post_recovery / self.reads_output

    @pydantic.computed_field(  # type: ignore
        description="The redundancy of node detection. The saturation is calculated as: `1 - # graph node / # graph reads`",
        return_type=float,
    )
    @property
    def node_saturation(self):
        """Return the node saturation."""
        return 1 - self.node_count_post_recovery / self.reads_output

    @pydantic.computed_field(  # type: ignore
        description="The number of reads discarded due to cell size filtering and multiplet recovery.",
        return_type=int,
    )
    @property
    def discarded_reads(self) -> int:
        """Return the total number of discarded reads in the graph stage."""
        return self.reads_input - self.reads_output

    @pydantic.computed_field(  # type: ignore
        description="The number of molecules discarded due to cell size filtering and multiplet recovery.",
        return_type=int,
    )
    @property
    def discarded_molecules(self) -> int:
        """Return the total number of discarded reads in the graph stage."""
        return self.molecules_input - self.molecules_output



"""Graph backend implementations used by pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from typing import Literal, Optional, Union

import networkx as nx

from pixelator.common.graph.backends.implementations._networkx import (
    NetworkXGraphBackend,
)
from pixelator.common.graph.backends.protocol import GraphBackend

__all__ = ["NetworkXGraphBackend"]

logger = logging.getLogger(__name__)


def graph_backend(
    graph_backend_class: Optional[Literal["NetworkXGraphBackend"]] = None,
) -> GraphBackend:
    """Get a graph backend.

    Pick up a GraphBackend. Defaults to `NetworkXGraphBackend`
    (at this time the only available backend).
    :param graph_backend_class: name of the graph backend class to try to pickup.
    :returns: A concrete graph backend instance
    :rtype: GraphBackend
    :raises ValueError: when `graph_backend_class` is not recognized
    """
    # TODO Later on we could use this as an entry point for loading
    # a graph backend from plugins if we would like.

    def _load_nx():
        logger.debug("Setting up a networkx based backend")
        return NetworkXGraphBackend

    if graph_backend_class:
        if graph_backend_class == "NetworkXGraphBackend":
            return _load_nx()

        raise ValueError(
            f"Class name {graph_backend_class} not recognized as `GraphBackend`"
        )

    return _load_nx()


def graph_backend_from_graph_type(graph: Union[nx.Graph, nx.MultiGraph]):
    """Pick the correct backend type based on the graph type."""
    if isinstance(graph, nx.Graph):
        return graph_backend("NetworkXGraphBackend")
    raise ValueError("Cannot recognize type of `graph`")



"""Helper commands for pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

from typing import Any

import click

from pixelator.common.utils import click_echo


def list_single_cell_designs(ctx: click.Context, param: Any, value: Any) -> None:
    """Return a list of single cell designs supported by the config.

    :param ctx: The click context
    :param param: The click parameter
    :param value: The click value
    """
    from pixelator.mpx.config import config

    if not value or ctx.resilient_parsing:
        return

    options = list(config.assays.keys())
    for option in options:
        click_echo(option)

    ctx.exit()


def list_single_cell_panels(ctx: click.Context, param: Any, value: Any) -> None:
    """Return a list of single cell panels supported by the config.

    :param ctx: The click context
    :param param: The click parameter
    :param value: The click value
    """
    from pixelator.mpx.config import config

    if not value or ctx.resilient_parsing:
        return

    options = config.list_panel_names(include_aliases=True)
    for option in options:
        click_echo(option)

    ctx.exit()



"""Model for report data returned by the single-cell adapterqc stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
from pathlib import Path

import pydantic

from pixelator.common.utils import get_sample_name
from pixelator.mpx.report.models.base import SampleReport


class AdapterQCSampleReport(SampleReport):
    """Model for data returned by the adapterqc stage."""

    total_read_count: int = pydantic.Field(
        ...,
        description="The total number of input reads in the adapterqc stage.",
    )

    passed_filter_read_count: int = pydantic.Field(
        ...,
        description="The number of reads that passed the filter in the adapterqc stage.",
    )

    @pydantic.computed_field(  # type: ignore
        return_type=float,
        description="The fraction of reads that was discarded in this stage.",
    )
    @property
    def discarded(self) -> float:  # noqa: D102
        return 1 - (self.passed_filter_read_count / self.total_read_count)

    @classmethod
    def from_json(cls, p: Path) -> AdapterQCSampleReport:
        """Initialize an :class:`AdapterQCSampleReport` from a report file."""
        sample_name = get_sample_name(p)

        with open(p) as fp:
            json_data = json.load(fp)

        data = {
            "total_read_count": json_data["read_counts"]["input"],
            "passed_filter_read_count": json_data["read_counts"]["output"],
        }

        return cls(sample_id=sample_name, **data)



"""
Console script for pixelator (analysis)

Copyright © 2022 Pixelgen Technologies AB.
"""

from typing import get_args

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.analysis import analyse_pixels
from pixelator.mpx.analysis.colocalization import ColocalizationAnalysis
from pixelator.mpx.analysis.colocalization.types import TransformationTypes
from pixelator.mpx.analysis.polarization import PolarizationAnalysis
from pixelator.mpx.cli.common import logger, output_option


@click.command(
    "analysis",
    short_help="compute different downstream analyses on annotated samples",
    options_metavar="<options>",
)
@click.argument(
    "pxl_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="PXL",
)
@click.option(
    "--compute-polarization",
    required=False,
    is_flag=True,
    help="Compute polarization scores matrix (components by markers)",
)
@click.option(
    "--compute-colocalization",
    required=False,
    is_flag=True,
    help="Compute colocalization scores (marker by marker) for each component",
)
@click.option(
    "--use-full-bipartite",
    required=False,
    is_flag=True,
    help=(
        "Use the bipartite graph instead of the one-node projection (UPIA)"
        " when computing polarization, coabundance and colocalization scores"
    ),
)
@click.option(
    "--polarization-transformation",
    default="log1p",
    required=False,
    type=click.Choice(["raw", "log1p"]),
    show_default=True,
    help=(
        "Which approach to use to normalize the antibody counts:"
        " \n\traw will use the raw counts"
        " \n\tlog1p will use the log(x+1) transformed counts"
    ),
)
@click.option(
    "--polarization-n-permutations",
    default=50,
    required=False,
    type=click.IntRange(min=5),
    show_default=True,
    help=(
        "Set the number of permutations use to compute the empirical"
        " z-score and p-value for the polarization score."
    ),
)
@click.option(
    "--polarization-min-marker-count",
    default=5,
    required=False,
    type=click.IntRange(min=2),
    show_default=True,
    help=(
        "Set the minimum number of counts of a marker to calculate"
        " the polarization score in a component."
    ),
)
@click.option(
    "--colocalization-transformation",
    default="rate-diff",
    required=False,
    type=click.Choice(get_args(TransformationTypes)),
    show_default=True,
    help=(
        "Select the type of transformation to use on the "
        "node by antibody counts matrix when computing colocalization"
    ),
)
@click.option(
    "--colocalization-neighbourhood-size",
    default=1,
    required=False,
    type=click.IntRange(min=0),
    show_default=True,
    help=(
        "Select the size of the neighborhood to use when computing "
        "colocalization metrics on each component"
    ),
)
@click.option(
    "--colocalization-n-permutations",
    default=50,
    required=False,
    type=click.IntRange(min=5),
    show_default=True,
    help=(
        "Set the number of permutations use to compute the empirical "
        "p-value for the colocalization score"
    ),
)
@click.option(
    "--colocalization-min-region-count",
    default=5,
    required=False,
    type=click.IntRange(min=0),
    show_default=True,
    help=(
        "The minimum number of counts in a region for it to be considered "
        "valid for computing colocalization"
    ),
)
@click.option(
    "--colocalization-min-marker-count",
    default=5,
    required=False,
    type=click.IntRange(min=0),
    show_default=True,
    help=("The minimum number of marker counts in component for colocalization"),
)
@output_option
@click.pass_context
@timer
def analysis(
    ctx,
    pxl_file,
    compute_polarization,
    compute_colocalization,
    use_full_bipartite,
    polarization_transformation,
    polarization_n_permutations,
    polarization_min_marker_count,
    colocalization_transformation,
    colocalization_neighbourhood_size,
    colocalization_n_permutations,
    colocalization_min_region_count,
    colocalization_min_marker_count,
    output,
):
    """
    Perform different analyses on a PixelDataset from pixelator annotate
    """
    # log input parameters
    input_files = [pxl_file]

    log_step_start(
        "analysis",
        input_files=input_files,
        output=output,
        compute_polarization=compute_polarization,
        compute_colocalization=compute_colocalization,
        polarization_transformation=polarization_transformation,
        polarization_n_permutations=polarization_n_permutations,
        polarization_min_marker_count=polarization_min_marker_count,
        colocalization_transformation=colocalization_transformation,
        colocalization_neighbourhood_size=colocalization_neighbourhood_size,
        colocalization_n_permutations=colocalization_n_permutations,
        colocalization_min_region_count=colocalization_min_region_count,
        colocalization_min_marker_count=colocalization_min_marker_count,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions="pxl")

    # sanity check on the anlyses
    if not any([compute_polarization, compute_colocalization]):
        logger.warning("All the analysis are disabled, no scores will be computed")

    # create output folder if it does not exist
    analysis_output = create_output_stage_dir(output, "analysis")

    # compute graph/clusters using parallel processing
    logger.info(f"Computing analysis for file {pxl_file}")

    clean_name = get_sample_name(pxl_file)
    metrics_file = analysis_output / f"{clean_name}.report.json"

    write_parameters_file(
        ctx,
        analysis_output / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx analysis",
    )

    analysis_to_run = []
    if compute_polarization:
        logger.info("Polarization score computation is activated")
        analysis_to_run.append(
            PolarizationAnalysis(
                transformation_type=polarization_transformation,
                n_permutations=polarization_n_permutations,
                min_marker_count=polarization_min_marker_count,
            )
        )

    if compute_colocalization:
        logger.info("Colocalization score computation is activated")
        analysis_to_run.append(
            ColocalizationAnalysis(
                transformation_type=colocalization_transformation,
                neighbourhood_size=colocalization_neighbourhood_size,
                n_permutations=colocalization_n_permutations,
                min_region_count=colocalization_min_region_count,
                min_marker_count=colocalization_min_marker_count,
            )
        )

    analyse_pixels(
        input=pxl_file,
        output=str(analysis_output),
        output_prefix=clean_name,
        metrics_file=str(metrics_file),
        use_full_bipartite=use_full_bipartite,
        analysis_to_run=analysis_to_run,
    )



"""Functions related to the graph dataclass used in pixelator graph operations.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
from functools import cached_property, lru_cache
from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

import pandas as pd
import polars as pl
from scipy.sparse import csr_array, csr_matrix

from pixelator.common.graph.backends.implementations import (
    graph_backend,
    graph_backend_from_graph_type,
)
from pixelator.common.graph.backends.protocol import (
    GraphBackend,
    SupportedLayoutAlgorithm,
    VertexClustering,
)
from pixelator.common.graph.node_metrics import local_g

logger = logging.getLogger(__name__)


class Graph:
    """`Graph` represents a graph, i.e. a collection of vertices and edges."""

    def __init__(self, backend: GraphBackend):
        """Create a new Graph instance.

        Create a Graph instance (as an end-user this is probably not the interface
        you are looking for). Try `Graph.from_edgelist`.

        :param backend: The backend used to represent the graph
        """
        self._backend = backend
        self._connected_components_needs_recompute = False
        self._connected_components: Optional[VertexClustering] = None

    @staticmethod
    def from_edgelist(
        edgelist: Union[pd.DataFrame, pl.LazyFrame],
        add_marker_counts: bool,
        simplify: bool,
        use_full_bipartite: bool,
        convert_indices_to_integers: bool = True,
    ) -> Graph:
        """Build a graph from an edgelist.

        Build a Graph from an edge list (pd.DataFrame). Multiple options are available
        to build the graph, `add_marker_counts` will add a dictionary of marker counts
        to each node, `simplify` will remove redundant edges and `use_full_bipartite`
        will not project the graph (UPIA).

        The graph will contain the edge attributes present in the edge list when
        `use_full_bipartite` is True and a dictionary of marker counts in each
        vertex (node) when `add_marker_counts` is True. If `use_full_bipartite` is
        False or `simplify` is True the edge attributes will be lost.

        :param edgelist: the edge list (dataframe) corresponding to the graph either as
                         a pandas data frame or as a polars LazyFrame. To minimize the
                         memory usage the LazyFrame is preferred.
        :param add_marker_counts: add a dictionary of marker counts to each node
        :param simplify: simplifies the graph (remove redundant edges)
        :param use_full_bipartite: use the bipartite graph instead of the projection
                                  (UPIA)
        :param convert_indices_to_integers: convert the indices to integers (this is the default)
        :returns: a Graph instance
        :rtype: Graph
        :raises: AssertionError when the input edge list is not valid
        """
        backend = graph_backend().from_edgelist(
            edgelist=edgelist,
            add_marker_counts=add_marker_counts,
            simplify=simplify,
            use_full_bipartite=use_full_bipartite,
            convert_indices_to_integers=convert_indices_to_integers,
        )

        return Graph(backend=backend)

    @staticmethod
    def from_raw(graph: Any) -> Graph:
        """Generate a Graph from a graph object.

        Which graph object is valid depends on the underlying
        graph implementation. In general end users should use
        the `from_edgelist` method instead.

        :param graph: input graph to use
        :return: A pixelator Graph object
        :rtype: Graph
        :raises ValueError: if type of `graph` is not recognized.
        """
        Backend = graph_backend_from_graph_type(graph)
        return Graph(backend=Backend(graph))

    @property
    def raw(self):
        """Get the underlying raw graph implementation.

        The type of this will depend on the underlying backend.
        This property is useful when you need to inter-operate
        with other graph libraries that work e.g. with a
        networkx graph instance.
        """
        return self._backend.raw

    @property
    def vs(self):
        """Get a sequence of the vertices in the Graph instance."""
        return self._backend.vs

    @property
    def es(self):
        """A sequence of the edges in the Graph instance."""
        return self._backend.es

    @cached_property
    def node_marker_counts(self):
        """Get the node marker counts as a dataframe."""
        return self._backend.node_marker_counts()

    def vcount(self):
        """Get the total number of vertices in the Graph instance."""
        return self._backend.vcount()

    def ecount(self):
        """Get the total number of edges in the Graph instance."""
        return self._backend.ecount()

    def get_adjacency_sparse(
        self, node_ordering: Iterable[Any] | None = None
    ) -> csr_matrix:
        """Get the sparse adjacency matrix.

        :param node_ordering: Control the node ordering in the adjacency matrix
        :return: a sparse adjacency matrix
        """
        return self._backend.get_adjacency_sparse(node_ordering=node_ordering)

    def connected_components(self) -> VertexClustering:
        """Get the connected components in the Graph instance."""
        # Recompute if there has been changes, otherwise pick the
        # cached value
        if not self._connected_components:
            self._connected_components = self._backend.connected_components()
            self._connected_components_needs_recompute = False
        if self._connected_components and self._connected_components_needs_recompute:
            self._connected_components = self._backend.connected_components()
            self._connected_components_needs_recompute = False

        return self._connected_components  # type: ignore

    def layout_coordinates(
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        only_keep_a_pixels: bool = True,
        get_node_marker_matrix: bool = True,
        cache: bool = False,
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate coordinates and (optionally) node marker counts for plotting.

        Generate a dataframe with coordinates, and (optionally) node marker
        counts to use that can be used for plotting.

        The layout options are:
          - fruchterman_reingold
          - fruchterman_reingold_3d
          - kamada_kawai
          - kamada_kawai_3d
          - pmds
          - pmds_3d
          - wpmds_3d


        The `pmds` options are much faster than the force-directed algorithms fruchterman_reingold
        and kamada_kawai. The `wpmds_3d` option is a weighted version of the `pmds_3d` algorithm.

        :param layout_algorithm: the layout algorithm to use to generate the coordinates
        :param only_keep_a_pixels: If true, only keep the a-pixels
        :param get_node_marker_matrix: Add a matrix of marker counts to each
                                       node if True.
        :param cache: set to `True` in order to cache one call of this this method.
                      It will make subsequent calls to the layout method
                      with the same settings much faster, at the cost of additional
                      memory usage. This can speed things up a lot when plotting
                      e.g. different markers across multiple markers.
        :param random_seed: used as the seed for graph layouts with a stochastic
                            element. Useful to get deterministic layouts across
                            method calls.
        :param **kwargs: will be passed to the underlying layout implementation
        :return: the coordinates and markers (if activated) as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if the provided `layout_algorithm` is not valid
        :raises: ValueError if the provided current graph instance is empty
        """
        if cache:
            return self._cached_layout_coordinates(
                layout_algorithm=layout_algorithm,
                only_keep_a_pixels=only_keep_a_pixels,
                get_node_marker_matrix=get_node_marker_matrix,
                random_seed=random_seed,
                **kwargs,
            )
        else:
            return self._backend.layout_coordinates(
                layout_algorithm=layout_algorithm,
                only_keep_a_pixels=only_keep_a_pixels,
                get_node_marker_matrix=get_node_marker_matrix,
                random_seed=random_seed,
                **kwargs,
            )

    @lru_cache(maxsize=1)
    def _cached_layout_coordinates(self, **kwargs):
        return self._backend.layout_coordinates(**kwargs)

    def get_edge_dataframe(self):
        """Get the edges as a pandas DataFrame."""
        return self._backend.get_edge_dataframe()

    def get_vertex_dataframe(self):
        """Get all vertices as a pandas DataFrame."""
        return self._backend.get_vertex_dataframe()

    def add_edges(self, edges: Iterable[Tuple[int]]) -> None:
        """Add edges to the graph instance.

        :param edges: Add the following edges to the graph instance.
        """
        self._backend.add_edges(edges)
        self._connected_components_needs_recompute = True

    def add_vertices(self, n_vertices: int, attrs: Dict[str, List]) -> None:
        """Add some number of vertices to the graph instance.

        :param n_vertices: the number of vertices to be added to the graph instance.
        :param attrs: dict of sequences, all of length equal to the number of vertices
                      to be added, containing the attributes of the new vertices. If
                      `n_vertices=1` then they have to be lists of length 1.
        :raises IndexError: if the number of graph vertices to add and lists of
                            attributes are of different lengths
        """
        self._backend.add_vertices(n_vertices=n_vertices, attrs=attrs)
        self._connected_components_needs_recompute = True

    def add_names_to_vertexes(self, vs_names: List[str]) -> None:
        """Rename the current vertices on the graph instance.

        :param vs_names: Add the following vertices to the graph instance.
        :raises ValueError: if the graph is empty
        :raises IndexError: if the number of graph vertices and list of names are
                            of different length
        """
        self._backend.add_names_to_vertexes(vs_names=vs_names)

    def __repr__(self) -> str:
        """Return a string representation of the graph."""
        return f"Graph with {self.vcount()} vertices and {self.ecount()} edges"

    def local_g(
        self,
        k: int = 1,
        use_weights: bool = True,
        normalize_counts: bool = True,
        W: csr_array | None = None,
        method: Literal["gi", "gstari"] = "gi",
    ) -> pd.DataFrame:
        """Compute the local G metric for each node in the graph.

        The local G metric is a measure of the local clustering of a node in the graph.

        :param k: The number of steps in the k-step random walk. Default is 1.
        :param use_weights: Whether to use weights in the computation. When turned off, all
        edge weights will be equal to 1. Default is True.
        :param normalize_counts: Whether to normalize counts to proportions. Default is True.
        :param W: A sparse matrix of custom edge weights. This will override the automated
        computation of edge weights. `W` must have the same dimensions as A. Note that weights can
        be defined for any pair of nodes, not only the pairs represented by edges in `A`. Default is None.
        :param method: The method to use for computing local G. Must be one of 'gi' or 'gstari'.
        'gi' is the original local G metric, which does not consider self-loops, meaning that the
        local marker expression for a node is computed by aggregating the weighted expression of
        its neighbors. 'gstari' is a simplified version of local G that does consider self-loops.
        In other words, the local marker expression of a node also includes the weighted marker
        expression of the node itself. Default is 'gi'.
        :return: A DataFrame of local G-scores for each node and marker.
        """
        counts = self.node_marker_counts
        A = self.get_adjacency_sparse(node_ordering=counts.index)
        return local_g(
            A=A,
            counts=counts,
            k=k,
            use_weights=use_weights,
            normalize_counts=normalize_counts,
            W=W,
            method=method,
        )



"""Constants for the collapse module.

This module contains constants for the collapse and error correction
module of MPX data (from FASTQ).

Copyright © 2023 Pixelgen Technologies AB.
"""

SEED = 42



"""PNA Pixel Dataset.

Copyright © 2025 Pixelgen Technologies AB.
"""

from __future__ import annotations

import copy
import tempfile
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Literal

import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
from anndata import AnnData, ImplicitModificationWarning

from pixelator.common.statistics import clr_transformation, log1p_transformation
from pixelator.pna.graph import PNAGraph, PNAGraphBackend
from pixelator.pna.pixeldataset.io import (
    InplacePixelDataFilterer,
    PixelDataQuerier,
    PixelDataViewer,
    PxlFile,
    copy_databases,
)
from pixelator.pna.utils import normalize_input_to_list, normalize_input_to_set


def read(paths: Path | list[Path] | str | list[str]) -> PNAPixelDataset:
    """Read a PNAPixelDataset from one or more provided .pxl file(s).

    :param path: path to the file to read
    :return: an instance of `PNAPixelDataset`
    """
    if not isinstance(paths, list):
        paths = [paths]  # type: ignore
    normalized_paths = [Path(p) for p in paths]  # type: ignore
    return PNAPixelDataset.from_pxl_files(normalized_paths)


@dataclass(slots=True, frozen=True, repr=True)
class Component:
    """A dataclass to hold a component and its associated graph."""

    component_id: str
    frame: pl.LazyFrame

    @property
    def graph(self) -> PNAGraph:
        """Get the graph."""
        return PNAGraph(PNAGraphBackend.from_edgelist(self.frame))


class Edgelist:
    """Representation of an edgelist.

    To get the edgelist as a pandas DataFrame, use the `to_df()` method.
    To get the edgelist as a polars DataFrame, use the `to_polars()` method.
    To access the components one at a time use the `iterator()` method.

    For memory efficient access to the edgelist, use the `to_record_batches()` method.
    This allows you to iterate over the edgelist using py.ArrowRecordBatch's.
    This is useful when you have a large edgelist that does not fit into memory,
    and your algorithm can be implemented in a streaming fashion.
    """

    def __init__(
        self,
        querier: PixelDataQuerier,
        components: str | Iterable[str] | None = None,
    ):
        """Create a new instance of Edgelist."""
        self._querier = querier
        self._components = normalize_input_to_set(components)

    @property
    def components(self) -> set[str]:
        """Get the component names."""
        return self._components or set(self._querier.read_all_component_names())

    def _handle_backwards_compatibility(
        self, df: pl.DataFrame | pd.DataFrame
    ) -> pl.DataFrame | pd.DataFrame:
        # Handle legacy marker names
        if isinstance(df, pl.DataFrame):
            return df.rename(
                {"marker1": "marker_1", "marker2": "marker_2"}, strict=False
            )
        else:
            return df.rename(columns={"marker1": "marker_1", "marker2": "marker_2"})

    def __len__(self) -> int:
        """Get the number of edges in the edgelist."""
        return self._querier.read_edgelist_len(components=self._components)

    def is_empty(self) -> bool:
        """Check if the edgelist is empty."""
        return len(self) == 0

    def to_df(self) -> pd.DataFrame:
        """Get the edgelist as a pandas DataFrame."""
        df = self._querier.read_edgelist(components=self.components, as_pandas=True)
        return self._handle_backwards_compatibility(df)

    def to_polars(self) -> pl.DataFrame:
        """Get the edgelist as a polars DataFrame."""
        # TODO change this once we can get filtering pushdown in duckdb
        # But for now this, somewhat counter-intuitively is the faster
        df = pl.concat([df for _, df in self._iterator()])
        return self._handle_backwards_compatibility(df)

    def to_record_batches(
        self, batch_size: int = 1_000_000
    ) -> Iterable[pa.RecordBatch]:
        """Get the edgelist as a stream of pyarrow RecordBatches."""
        return self._querier.read_edgelist_stream(
            components=self.components, batch_size=batch_size
        )

    def _iterator(self) -> Iterable[tuple[str, pl.DataFrame]]:
        for component in self.components:
            yield component, self._querier.read_edgelist(components=component)

    def iterator(self) -> Iterable[Component]:
        """Get a stream of components and their graphs.

        :return: A stream of component names and associated graphs
        """
        for name, df in self._iterator():
            yield Component(
                component_id=name, frame=self._handle_backwards_compatibility(df).lazy()
            )

    def __str__(self) -> str:
        """Get a string representation of the Edgelist."""
        n_edges = len(self)
        n_components = len(self.components)
        if n_components <= 5:
            return f"EdgeList({n_edges:,} edges in component set: {self.components})"
        return f"EdgeList({n_edges:,} edges in {n_components} components)"

    def __repr__(self):
        """Get a string representation of the Edgelist."""
        return str(self)

    def _ipython_display_(self):
        """Display the Edgelist in Jupyter notebooks."""
        return print(str(self))


class Proximity:
    """Representation of a proximity data.

    This class allows you to access the proximity data as a pandas or polars
    DataFrame. From there you can analyze the data further using your favorite
    data analysis tools.
    """

    def __init__(
        self,
        querier: PixelDataQuerier,
        components: str | list[str] | set[str] | None = None,
        markers: str | list[str] | set[str] | None = None,
        add_marker_counts: bool = True,
        add_log2_ratio: bool = True,
    ):
        """Create a new instance of Proximity."""
        self._querier = querier
        self._components = normalize_input_to_set(components)
        self._markers = normalize_input_to_set(markers)
        self._add_marker_counts = add_marker_counts
        self._add_log2_ratio_col = add_log2_ratio

    @property
    def components(self) -> set[str]:
        """Get the component names."""
        return (
            self._components
            if self._components is not None
            else set(self._querier.read_all_component_names())
        )

    @property
    def markers(self) -> set[str]:
        """Get the marker names."""
        return (
            self._markers
            if self._markers is not None
            else set(self._querier.read_all_marker_names())
        )

    def __len__(self) -> int:
        """Get the number of proximity scores."""
        return self._querier.read_proximity_len(
            components=self._components, markers=self._markers
        )

    def is_empty(self):
        """Check if the proximity data is empty."""
        return len(self) == 0

    def _add_marker_counts_to_proximity_df(
        self, adata: AnnData, proximity_df: pl.DataFrame
    ):
        marker_counts = pl.DataFrame(adata.to_df().reset_index())
        node_counts = (
            marker_counts.drop("component")
            .transpose(column_names=marker_counts["component"])
            .sum()
            .transpose(column_names=["node_counts"])
            .with_columns(component=marker_counts["component"])
        )
        marker_counts = marker_counts.unpivot(
            index="component", variable_name="marker", value_name="marker_1_count"
        )
        marker_counts = marker_counts.with_columns(
            pl.col("marker_1_count").alias("marker_2_count")
        )
        # Cast to uint32 integers to lower memory usage
        marker_counts = marker_counts.with_columns(
            marker_1_count=pl.col("marker_1_count").cast(pl.UInt32),
            marker_2_count=pl.col("marker_2_count").cast(pl.UInt32),
        )
        marker_counts = marker_counts.join(
            node_counts,
            on="component",
            how="left",
        )
        marker_counts = marker_counts.with_columns(
            marker_1_freq=pl.col("marker_1_count") / pl.col("node_counts"),
            marker_2_freq=pl.col("marker_2_count") / pl.col("node_counts"),
        )
        proximity_df = (
            proximity_df.join(
                marker_counts.select(
                    ["component", "marker", "marker_1_count", "marker_1_freq"]
                ),
                left_on=["component", "marker_1"],
                right_on=["component", "marker"],
                how="left",
            )
            .join(
                marker_counts.select(
                    ["component", "marker", "marker_2_count", "marker_2_freq"]
                ),
                left_on=["component", "marker_2"],
                right_on=["component", "marker"],
                how="left",
            )
            .with_columns(
                min_count=pl.min_horizontal("marker_1_count", "marker_2_count")
            )
        )

        return proximity_df

    def _add_log2_ratio(self, proximity):
        return proximity.with_columns(
            log2_ratio=np.log2(
                np.maximum(pl.col("join_count"), 1)
                / np.maximum(pl.col("join_count_expected_mean"), 1)
            )  # setting values <1 to 1 to avoid division by zero
        )

    def _post_process(self, df: pl.DataFrame) -> pl.DataFrame:
        if self._add_marker_counts:
            adata = self._querier.read_adata()
            df = self._add_marker_counts_to_proximity_df(adata, df)

        if self._add_log2_ratio_col:
            df = self._add_log2_ratio(df)

        return df

    def to_df(self) -> pd.DataFrame:
        """Get the edgelist as a pandas DataFrame."""
        return self.to_polars().to_pandas()

    def to_polars(self) -> pl.DataFrame:
        """Get the edgelist as a polars DataFrame."""
        return self._post_process(
            self._querier.read_proximity(
                components=self._components, markers=self._markers
            )
        )

    def __str__(self) -> str:
        """Get a string representation of the Proximity."""
        n_proximity = len(self)
        return ", ".join(
            [
                f"Proximity({n_proximity:,} elements",
                f"add_marker_counts={self._add_marker_counts}",
                f"add_logratio={self._add_log2_ratio_col})",
            ]
        )

    def __repr__(self):
        """Get a string representation of the Proximity."""
        return str(self)

    def _ipython_display_(self):
        return print(str(self))


class PreComputedLayouts:
    """Representation of precomputed layouts.

    This contains precomputed layouts for one or more components.
    """

    def __init__(
        self,
        querier: PixelDataQuerier,
        components: str | Iterable[str] | None = None,
        add_marker_counts: bool = True,
        add_spherical_norm: bool = False,
    ):
        """Create a new instance of PreComputedLayouts."""
        self._querier = querier
        self._components = normalize_input_to_set(components)
        self._add_marker_counts = add_marker_counts
        self._add_spherical_norm = add_spherical_norm

    @property
    def components(self) -> set[str]:
        """Get the component names."""
        return (
            self._components
            if self._components is not None
            else set(self._querier.read_all_component_names())
        )

    def __len__(self) -> int:
        """Get the nodes in the layouts."""
        return self._querier.read_layouts_len(components=self._components)

    def is_empty(self) -> bool:
        """Check if the precomputed layouts are empty."""
        return len(self) == 0

    def to_df(self) -> pd.DataFrame:
        """Get the precomputed layouts as a pandas DataFrame."""
        return self.to_polars().to_pandas()

    def _handle_backwards_comp(self, df: pl.DataFrame) -> pl.DataFrame:
        # If the norm data is in the output (which it is for some legacy files)
        # drop them.
        return df.drop(["x_norm", "y_norm", "z_norm"], strict=False)

    def _post_process(self, df: pl.DataFrame) -> pl.DataFrame:
        df = self._handle_backwards_comp(df)

        if self._add_spherical_norm:
            coordinates = df.select(["x", "y", "z"]).to_numpy()
            normalized_coordinates = pl.DataFrame(
                coordinates / (1 * np.linalg.norm(coordinates, axis=1))[:, None],
                schema={
                    "x_norm": pl.Float32,
                    "y_norm": pl.Float32,
                    "z_norm": pl.Float32,
                },
            )
            df = df.hstack(normalized_coordinates)

        return df

    def to_polars(self) -> pl.DataFrame:
        """Get the precomputed layouts as a polars DataFrame."""
        return self._post_process(
            self._querier.read_layouts(
                components=self._components, add_marker_counts=self._add_marker_counts
            )
        )

    def iterator(
        self, return_polars_df: bool = False
    ) -> Iterable[tuple[str, pd.DataFrame | pl.DataFrame]]:
        """Get a stream of layouts.

        Control the trade-off between memory usage and performance by setting
        the `batch_size` parameter. If you set it to a large number, you will
        use more memory, but the performance should overall be better.

        :param return_polars_df: If True, return polars DataFrames, otherwise return pandas DataFrames

        :return: A stream of layouts names and associated layout dataframes
        """
        for component in self.components:
            component_df = self._post_process(
                self._querier.read_layouts(
                    components=component, add_marker_counts=self._add_marker_counts
                )
            )
            if return_polars_df:
                yield (component, component_df)
            else:
                yield (component, component_df.to_pandas())

    def describe(self) -> str:
        """Return a description of the PreComputedLayouts."""
        return f"PreComputedLayouts({len(self.components):,} components, {len(self):,} datapoints)"

    def __str__(self):
        """Get a string representation of the PreComputedLayouts."""
        return f"PreComputedLayouts({len(self.components):,} components)"

    def __repr__(self):
        """Get a string representation of the PreComputedLayouts."""
        return str(self)

    def _ipython_display_(self):
        return print(self.describe())


@dataclass
class PixelDatasetConfig:
    """Configuration for a PixelDataset."""

    adata_join_method: Literal["inner", "outer"] = "inner"


class PNAPixelDataset:
    """A PixelDataset is a collection of samples, components, and markers.

    This class provides a high-level interface to the data stored in one or more .pxl files.
    You can build a PixelDataset from one or more .pxl files, and then use the various methods
    to filer and access the underlying data in different ways.

    .. code-block:: python
        from pathlib import Path
        from pixelator.pna.pixeldataset import PixelDataset

        pxl_files = Path("<dir with pxl files>").glob("*.pxl")
        pxl_dataset = PixelDataset.from_pxl_files(pxl_files)

    To filter data you can do:
    .. code-block:: python
        ten_components = pxl_dataset.adata.obs.index[:10]
        pxl_dataset.filter(components=ten_components)
    """

    def __init__(
        self,
        view: PixelDataViewer,
        config: PixelDatasetConfig | None = None,
        active_samples: Iterable[str] | str | None = None,
        active_components: Iterable[str] | str | None = None,
        active_markers: Iterable[str] | str | None = None,
    ):
        """Create a new PixelDataset instance.

        Note that setting any of the `active_*` parameters to None will include
        all samples, components, or markers.

        :param view: The PixelDataViewer instance to use for accessing the data.
        :param config: The configuration for the dataset.
        :param active_samples: The samples to include in the dataset.
        :param active_components: The components to include in the dataset.
        :param active_markers: The markers to include in the dataset.
        """
        self._view = view
        if config is None:
            config = PixelDatasetConfig()
        self._config = config

        self._active_samples = (
            normalize_input_to_set(active_samples)
            if active_samples
            else set(self._view.sample_names())
        )
        self._active_components = normalize_input_to_set(active_components)
        self._active_markers = normalize_input_to_set(active_markers)

    @staticmethod
    def from_files(
        pxl_files: Path
        | Iterable[Path]
        | Iterable[PxlFile]
        | PxlFile
        | dict[str, Path],
        config: PixelDatasetConfig | None = None,
    ) -> PNAPixelDataset:
        """Alias for `from_pxl_files`."""
        return PNAPixelDataset.from_pxl_files(pxl_files, config)

    @staticmethod
    def from_pxl_files(
        pxl_files: Path
        | Iterable[Path]
        | Iterable[PxlFile]
        | PxlFile
        | dict[str, Path],
        config: PixelDatasetConfig | None = None,
    ) -> PNAPixelDataset:
        """Create a new PixelDataset from one or more .pxl files.

        If you pass a list of .pxl files the name of the samples
        will be inferred from the sample name in the file.

        If you pass a dictionary of .pxl files the keys will be used as the sample names.

        :param pxl_files: The .pxl files to include in the dataset.
                          Can be a list of paths or a dictionary with sample names
                          as keys and paths as values.
        :param config: The configuration for the dataset.
        """
        if isinstance(pxl_files, Path):
            return PNAPixelDataset(
                PixelDataViewer.from_files(pxl_files=[PxlFile(pxl_files)]),
                config=config,
            )

        if isinstance(pxl_files, dict):
            return PNAPixelDataset(
                PixelDataViewer.from_sample_to_file_mappings(
                    {k: PxlFile(v) for k, v in pxl_files.items()}
                ),
                config=config,
            )

        if isinstance(pxl_files, PxlFile):
            return PNAPixelDataset(
                PixelDataViewer.from_files(pxl_files=[pxl_files]),
                config=config,
            )

        pxl_files = list(pxl_files)  # type: ignore
        # Either you are a PxlFile, then let's go!
        if all(isinstance(f, PxlFile) for f in pxl_files):
            return PNAPixelDataset(
                view=PixelDataViewer.from_files(pxl_files),  # type: ignore
                config=config,
            )

        # Or we will assume you are path and try that!
        return PNAPixelDataset(
            view=PixelDataViewer.from_files(pxl_files=[PxlFile(f) for f in pxl_files]),  # type: ignore
            config=config,
        )

    def sample_names(self) -> set[str]:
        """Return the set of sample names in the project."""
        return set(self._active_samples)  # type: ignore

    def components(self) -> set[str]:
        """Return the set of component names in the project."""
        return set(self.adata().obs.index.to_list())

    def markers(self) -> set[str]:
        """Return the set of marker names in the project."""
        return set(self.adata().var.index.to_list())

    @property
    def view(self) -> PixelDataViewer:
        """Return the PixelDataViewer instance used by the dataset.

        This can be used to write custom queries to the underlying data, using
        the duckdb connection API.

        You can find more information about the duckdb API here:
        https://duckdb.org/docs/api/python/overview

        Typically you do not need to bother with using the connection
        directly, but for certain advanced use cases it can boost performance
        by quite a bit.

        .. code-block:: python
            from pixelator.pna.pixeldataset import PixelDataset

            pxl_files = ...
            pxl_dataset = PixelDataset.from_pxl_files(pxl_files)
            with pxl_dataset.view as connection:
                df = connection.sql("SELECT * FROM edgelist WHERE markers = 'CD3'").to_df()

        :return: The PixelDataViewer instance used by the dataset.
        """
        return self._view

    def adata(
        self,
        add_log1p_transform: bool = True,
        add_clr_transform: bool = True,
    ) -> AnnData:
        """Return the AnnData instance for the dataset.

        This will be filtered to only include the active samples, components, and markers.
        :param add_log1p_transform: If True, add the log1p transformation to the data.
        :param add_clr_transform: If True, add the clr transformation to the data.
        :return: The AnnData instance for the dataset.
        """
        # TODO For performance reasons we might want to push down
        # the component and marker filtering to the individual datasets
        # but for now this should probably be fine
        adata = self._view.read_adata()
        if self._active_components:
            adata = adata[normalize_input_to_list(self._active_components), :]
        if self._active_markers:
            adata = adata[:, normalize_input_to_list(self._active_markers)]

        # add normalization layers if requested
        if [add_clr_transform, add_log1p_transform].count(True) > 1:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=ImplicitModificationWarning)
                counts_df = adata.to_df()
                if add_clr_transform:
                    counts_df_clr = clr_transformation(df=counts_df, axis=1)
                    counts_df_clr.index = counts_df_clr.index.astype(str)
                    adata.obsm["clr"] = counts_df_clr

                if add_log1p_transform:
                    counts_df_log1p = log1p_transformation(df=counts_df)
                    counts_df_log1p.index = counts_df_log1p.index.astype(str)
                    adata.obsm["log1p"] = counts_df_log1p

        return adata.copy()

    def edgelist(
        self,
    ) -> Edgelist:
        """Return the Edgelist instance for the dataset.

        This will be filtered to only include the active samples and components.
        :return: The Edgelist instance for the dataset.
        """
        return Edgelist(
            PixelDataQuerier(self.view),
            components=self._active_components,
        )

    def proximity(
        self,
        add_marker_counts: bool = True,
        add_logratio: bool = True,
    ) -> Proximity:
        """Return the Proximity instance for the dataset.

        This will be filtered to only include the active samples, components, and markers.

        :param add_marker_counts: If True, add the marker counts to the proximity data.
        :param add_logratio: If True, add the logratio to the proximity data.
        :return: The Proximity instance for the dataset.
        """
        return Proximity(
            PixelDataQuerier(self.view),
            components=self._active_components,
            markers=self._active_markers,
            add_marker_counts=add_marker_counts,
            add_log2_ratio=add_logratio,
        )

    def precomputed_layouts(
        self, add_marker_counts: bool = True, add_spherical_norm: bool = False
    ) -> PreComputedLayouts:
        """Return the PreComputedLayouts instance for the dataset.

        :param add_marker_counts: If True, add the marker counts to the precomputed layouts.
        :param add_spherical_norm: If True, add spherical coordinates to dataframe
        This will be filtered to only include the active samples and components.
        :return: The PreComputedLayouts instance for the dataset.
        """
        return PreComputedLayouts(
            PixelDataQuerier(self.view),
            components=self._active_components,
            add_marker_counts=add_marker_counts,
            add_spherical_norm=add_spherical_norm,
        )

    def metadata(
        self,
    ) -> dict:
        """Return the metadata for the dataset."""
        return PixelDataQuerier(self.view).read_metadata()

    @staticmethod
    def _copy_or_none(values_or_none):
        if values_or_none is None:
            return None
        return values_or_none.copy()

    def filter(
        self,
        samples: Iterable[str] | str | None = None,
        components: Iterable[str] | str | None = None,
        markers: Iterable[str] | str | None = None,
    ) -> PNAPixelDataset:
        """Filter the dataset to only include the specified samples, components, and markers.

        Filtering by components will apply to all data modalities (i.e. adata, edgelist, proximity, and precomputed layouts).
        However, filtering by markers will only apply to the adata and proximity data modalities, since filtering
        by markers in the edgelist and precomputed layouts will cause components to break up.

        Note that filtering is done lazily, so creating new filters is cheap. The actual filtering will only be done
        once the underlying data is accessed.

        :param samples: The samples to include in the dataset (default: None means no filter is applied).
        :param components: The components to include in the dataset (default: None means no filter is applied).
        :param markers: The markers to include in the dataset (default: None means no filter is applied).
        :raises ValueError: if all of the specified samples, components, or markers do not exist in the dataset.
        :return: A new PixelDataset with the specified samples, components, and markers
        """
        samples = normalize_input_to_set(samples)
        components = normalize_input_to_set(components)
        markers = normalize_input_to_set(markers)

        errors = []
        if samples and not samples.issubset(self.sample_names()):
            errors.append(
                "One or more of the specified samples do not exist in the dataset."
            )

        if components and not components.issubset(self.components()):
            errors.append(
                "One or more of the specified components do not exist in the dataset."
            )

        if markers and not markers.issubset(self.markers()):
            errors.append(
                "One or more of the specified markers do not exist in the dataset."
            )

        if errors:
            message = ["Failed to filter, for the following reasons: "]
            message.extend(errors)
            raise ValueError("\n".join(message))

        active_samples = samples or self._copy_or_none(self._active_samples)
        active_components = components or self._copy_or_none(self._active_components)
        active_markers = markers or self._copy_or_none(self._active_markers)

        return PNAPixelDataset(
            view=self._view,
            config=copy.copy(self._config),
            active_samples=active_samples,
            active_components=active_components,
            active_markers=active_markers,
        )

    def __repr__(self) -> str:
        """Return a string representation of the PixelDataset."""
        return str(self)

    def __str__(self) -> str:
        """Return a string representation of the PixelDataset."""
        return f"""PixelatorProject(with {len(self.sample_names())} samples)"""

    def _ipython_display_(self):
        """Display the PixelDataset in Jupyter notebooks."""
        return print(self.describe())

    def describe(self) -> str:
        """Return a description of the PixelDataset."""
        description = [f"""PixelDataset with {len(self.sample_names())} samples"""]
        description.append("")
        description += [f"""Mapping the following samples to files:"""]
        for sample_name, file in self.view.sample_to_file_mappings.items():
            description.append(f"\tSample: {sample_name}, File: {file}")

        description.append("")
        adata = self.adata()
        description += [f"""In total it contains:"""]
        description.append(f"{len(adata.obs)} components, {len(adata.var)} markers")
        return "\n".join(description)


class PixelDatasetSaver:
    """A class to save a PixelDataset to disk."""

    def __init__(self, pxl_dataset: PNAPixelDataset):
        """Create a new PixelDatasetSaver instance."""
        self.pxl_dataset = pxl_dataset

    def save(
        self,
        sample_name: str,
        output_path: Path | str,
        optimize_disk_usage: bool = True,
    ) -> PxlFile:
        """Save a sample from a the PixelDataset to disk as a single pxl file with any component filters applied to it.

        NB: for the time being, no marker filters are applied to the saved file.

        This will copy the entire sample to a new file, applying any filters that have been set on the PxlFile
        on-disk.

        :param sample_name: The name of the sample to save.
        :param output_path: The path to save the sample to.
        :param optimize_disk_usage: If True, the saved file will be optimized for disk usage. If this is active
                                    a temporary file will be written before the final file is written to disk.

        :return: The PxlFile pointing to the saved PixelDataset.
        """
        try:
            input_sample = self.pxl_dataset.view.sample_to_file_mappings[sample_name]
        except KeyError:
            raise ValueError(
                f"Sample {sample_name} not found in the PixelDataset. Use one of: {self.pxl_dataset.sample_names()}"
            )

        if isinstance(output_path, str):
            output_path = Path(output_path)

        input_sample_pxl_file = PxlFile(input_sample)

        if optimize_disk_usage:
            with tempfile.NamedTemporaryFile() as temp_file:
                tmp_output_sample = PxlFile.copy_pxl_file(
                    input_sample_pxl_file, Path(temp_file.name)
                )
                InplacePixelDataFilterer(tmp_output_sample).filter_components(
                    self.pxl_dataset.components(),
                    metadata=input_sample_pxl_file.metadata(),
                )
                copy_databases(tmp_output_sample.path, output_path)
                return PxlFile(output_path, sample_name)

        output_sample = PxlFile.copy_pxl_file(input_sample_pxl_file, output_path)
        InplacePixelDataFilterer(output_sample).filter_components(
            self.pxl_dataset.components(),
            metadata=input_sample_pxl_file.metadata(),
        )
        return output_sample



"""Functions for the analysis of PixelDataset.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from dataclasses import asdict, dataclass
from pathlib import Path

from pixelator.mpx.analysis.analysis_engine import PerComponentAnalysis, run_analysis
from pixelator.mpx.analysis.colocalization.types import TransformationTypes
from pixelator.mpx.analysis.polarization.types import PolarizationTransformationTypes
from pixelator.mpx.pixeldataset import (
    PixelDataset,
)
from pixelator.mpx.report.models import AnalysisSampleReport
from pixelator.mpx.report.models.analysis import (
    ColocalizationReport,
    PolarizationReport,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class AnalysisParameters:
    """Analysis parameters."""

    compute_polarization: bool
    compute_colocalization: bool
    use_full_bipartite: bool
    polarization_transformation: PolarizationTransformationTypes
    polarization_n_permutations: int
    polarization_min_marker_count: int
    colocalization_transformation: TransformationTypes
    colocalization_neighbourhood_size: int
    colocalization_n_permutations: int
    colocalization_min_region_count: int
    colocalization_min_marker_count: int


def analyse_pixels(
    input: str,
    output: str,
    output_prefix: str,
    metrics_file: str,
    use_full_bipartite: bool,
    analysis_to_run: list[PerComponentAnalysis],
) -> None:
    """Run analysis functions on a PixelDataset.

    This function takes a pxl file that has been generated
    with `pixelator annotate`. The function then uses the `edge list` and
    the `AnnData` to compute the scores (polarization, co-abundance and
    co-localization) which are then added to the `PixelDataset` (depending
    on which scores are enabled).

    :param input: the path to the PixelDataset (zip)
    :param output: the path to the output file
    :param output_prefix: the prefix to prepend to the output file
    :param metrics_file: the path to a JSON file to write metrics
    :param use_full_bipartite: use the bipartite graph instead of the
                               one-node-projection (UPIA)
    :param analysis_to_run: a list of analysis functions (`PerComponentAnalysis` instances) to apply
                            to each component
    :param verbose: run if verbose mode when true
    :returns: None
    :rtype: None
    :raises AssertionError: the input arguments are not valid
    """
    logger.debug("Parsing PixelDataset from %s", input)

    # load the PixelDataset
    dataset = PixelDataset.from_file(input)

    names_of_analyses = {analysis.ANALYSIS_NAME for analysis in analysis_to_run}

    compute_polarization = "yes" if "polarization" in names_of_analyses else "no"
    compute_colocalization = "yes" if "colocalization" in names_of_analyses else "no"

    metrics = dict()
    metrics["polarization"] = "yes" if compute_polarization else "no"
    metrics["colocalization"] = "yes" if compute_colocalization else "no"

    dataset = run_analysis(
        pxl_dataset=dataset,
        analysis_to_run=analysis_to_run,
        use_full_bipartite=use_full_bipartite,
    )

    dataset.metadata["analysis"] = {
        "params": {
            analysis.ANALYSIS_NAME: analysis.parameters()
            for analysis in analysis_to_run
        }
    }
    # save dataset
    dataset.save(
        Path(output) / f"{output_prefix}.analysis.dataset.pxl", force_overwrite=True
    )

    polarization_report = PolarizationReport()
    colocalization_report = ColocalizationReport()

    report = AnalysisSampleReport(
        sample_id=output_prefix,
        polarization=polarization_report if compute_polarization else None,
        colocalization=colocalization_report if compute_colocalization else None,
    )
    report.write_json_file(metrics_file, indent=4)
